import torch
import functools

def _calculate_projection_acc(
    projected_gt:       torch.Tensor,
    projected_est:      torch.Tensor,
    threshold:          float
):
    """
        We accept  a  pose estimation as correct when the 2D projection error is smaller than 
        a predefined threshold.
        Input: 
        predicted pxls based on the estimated pose and gt pixels derived from gt pose.
    """
    b, _, __ = projected_gt.shape
    e = torch.linalg.norm(projected_est - projected_gt, ord=2, dim=-1).mean(dim=-1)
    valid_invalid =  torch.where( e <  threshold, 1 , 0)
    valid = torch.count_nonzero(valid_invalid)
    acc = 100.0 * float(valid) / b
    return torch.tensor(acc).float()

class Projection(torch.nn.Module):
    def __init__(self,
        threshold: int = 5,
    ):
        super(Projection, self).__init__()
        self.threshold = threshold

    def forward(self,
        gt:         torch.Tensor,
        pred:       torch.Tensor,
    ) -> torch.Tensor:
        return _calculate_projection_acc(gt, pred, self.threshold)

Projection2 = functools.partial(
    Projection,
    threshold=2
)

Projection5 = functools.partial(
    Projection, 
    threshold=5
)

Projection10 = functools.partial(
    Projection, 
    threshold=10
)


import moai.utils.engine as mieng

import torch
import omegaconf.omegaconf
import typing
import logging
import inspect
import itertools

log = logging.getLogger(__name__)

__all__ = ['Metric']

class Metric(mieng.Single):
    def __init__(self,
        metrics: omegaconf.DictConfig,
        **kwargs: typing.Mapping[str, typing.Any],
    ):
        super(Metric, self).__init__(
            items=metrics, 
            name="metric",
        )
        loop = ((key, params) for key, params in kwargs.items() if hasattr(metrics, key))
        for k, p in loop:
            last_module = self.metric
            sig = inspect.signature(last_module.forward)
            for keys in zip(*list(p[prop] for prop in itertools.chain(sig.parameters, ['out']))):
                self.execs.append(lambda tensor_dict, metric_dict, k=keys, p=sig.parameters.keys(), f=last_module:
                    metric_dict.update({
                        k[-1]: f(**dict(zip(p, 
                            list(tensor_dict[i] for i in k[:-1])
                        )))
                    })
                )

    def forward(self,
        tensors: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        metrics = { }        
        for exe in self.execs:
            exe(tensors, metrics)
        returned = { }
        for k, m in metrics.items():
            returned[k] = torch.mean(m) if len(m.size()) > 0 else m                
        return returned


import os
import pytorch_lightning

class NoOp(pytorch_lightning.callbacks.ModelCheckpoint):
    def __init__(self, **kwargs):
        super(NoOp, self).__init__(
            filepath=os.getcwd(),
            monitor="",
            verbose=False,
            save_top_k=0,
            period=0,
            save_weights_only=False,
            mode='auto'
        )

    def _save_model(self, filepath: str):
        pass

    def on_validation_end(self, trainer, pl_module):
        pass



__SYMBOL_MAP__ = {
    # '!': '%21',
    '#': '%23',
    '$': '%24',
    '%': '%25',
    '&': '%26',
    '\'': '%27',
    '(': '%28',
    ')': '%29',
    '*': '%2A',
    '+': '%2B',
    ',': '%2C',
    '/': '%2F',
    ':': '%3A',
    ';': '%3B',
    '=': '%3D',
    '?': '%3F',
    '@': '%40',
    '[': '%5B',
    ']': '%5D',
    ' ': '%20',

    '<': '%3C',
    '>': '%3E',
    '\\': '%5C',
    '{': '%7B',
    '}': '%7D',
}

__SIZE_MAP__ = {
    'large': '%5Clarge',
    'vlarge': '%5CLarge',
    'vvlarge': '%5CLARGE',
    'huge': '%5Chuge',
    'vhuge': '%5CHuge',
}

#NOTE: https://gist.github.com/a-rodin/fef3f543412d6e1ec5b6cf55bf197d7b

if __name__ == "__main__":
    wing_loss = r"\begin{equation} wing(x) = \left\{ \begin{array}{ll}w \ln (1 + |x|/\epsilon)  & \text{if } |x| < w \\|x| - C  & \text{otherwise}\end{array}\right.\end{equation}"

    adaptive_wing_loss = r"\begin{equation} \small AWing(y,\!\hat{y})\! =\! \begin{cases}  \omega\! \ln(1\! +\!  \displaystyle |\frac{y\!-\!\hat{y}\!}{\epsilon}|^{\alpha-y})\! &\! \text{if } |(y\!-\!\hat{y})|\! <\! \theta   \\   A|y-\hat{y}\!| - C & \text{otherwise} \end{cases} \end{equation}"

    soft_wing_loss = r"\begin{equation}\mathrm{Wing}(x)=\left\{ \begin{array}{ll} \omega \ln({1+\frac{|x|}{\epsilon}})& \mathrm{if}\ |x| < \omega \\ |x| - C  &\mathrm{otherwise} \end{array} \right. \end{equation}"

    berhu_loss = r"\begin{equation}\mathcal{B}(x) = \begin{cases} |x| & |x| \leq c, \\ \frac{x^2 + c^2}{2c} & |x| > c. \\ \end{cases} \end{equation}"

    berhu_threshold = r"c = \frac{1}{5} \max_i(|\tilde{y}_i - y_i|)"

    std_kl_loss = r"\begin{equation}\mathrm{StandardKL}(\mu,\sigma) = \frac{1}{2} \displaystyle\sum_{i} (1+\log(\sigma_{i}^{2}) -\mu_{i}^{2}-\sigma_{i}^{2})\end{equation}"

    std_kl_beta_loss = r"\begin{equation}\beta\mathrm{-StandardKL}(\mu,\sigma) = \beta \, \, \frac{1}{2} \displaystyle\sum_{i} (1+\log(\sigma_{i}^{2}) -\mu_{i}^{2}-\sigma_{i}^{2})\end{equation}"

    std_kl_capacity_loss = r"\begin{equation}\mathrm{CapacityStandardKL}(\mu,\sigma) = \beta \, \, \, |\,\frac{1}{2} \displaystyle\sum_{i} (1+\log(\sigma_{i}^{2}) -\mu_{i}^{2}-\sigma_{i}^{2})\,-\,C\,|\end{equation}"

    std_kl_robust_loss = r"\begin{equation}\mathrm{RobustStandardKL}(\mu,\sigma) = \sqrt{1 + \big(\frac{1}{2} \displaystyle\sum_{i} (1+\log(\sigma_{i}^{2}) -\mu_{i}^{2}-\sigma_{i}^{2})\big)^2} - 1\end{equation}"

    geodesic_loss = r"\begin{equation}\mathcal{d}(R_1,R_2) = \arccos\frac{trace(R_1R_2^T) - 1}{2}\end{equation}"

    output_string = geodesic_loss

    size = 'huge'

    for k, v in __SYMBOL_MAP__.items():
        output_string = output_string.replace(k, v)

    url = "https://render.githubusercontent.com/render/math?math="
    print(url + __SIZE_MAP__[size] + output_string)

    


import torch
import math
import typing

__all__ = ['AdaBound', 'AdaBoundW']

#NOTE: modified from https://github.com/Luolc/AdaBound

class AdaBound(torch.optim.Optimizer):
    """Implements AdaBound algorithm.

    - **Paper**: [Adaptive Gradient Methods with Dynamic Bound of Learning Rate](https://arxiv.org/pdf/1902.09843.pdf).
    - **Implementation**: [GitHub Repo @ Luolc](https://github.com/Luolc/AdaBound)
    
    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): Adam learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        final_lr (float, optional): final (SGD) learning rate (default: 0.1)
        gamma (float, optional): convergence speed of the bound functions (default: 1e-3)
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsbound (boolean, optional): whether to use the AMSBound variant of this algorithm
    .. Adaptive Gradient Methods with Dynamic Bound of Learning Rate:
        https://openreview.net/forum?id=Bkg3g2R9FX
    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        final_lr: float=0.1,
        gamma: float=1e-3,
        eps: float=1e-8,
        weight_decay: float=0,
        amsbound: bool=False,
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= final_lr:
            raise ValueError("Invalid final learning rate: {}".format(final_lr))
        if not 0.0 <= gamma < 1.0:
            raise ValueError("Invalid gamma parameter: {}".format(gamma))
        defaults = dict(lr=lr, betas=betas, final_lr=final_lr, gamma=gamma, eps=eps,
                        weight_decay=weight_decay, amsbound=amsbound)
        super(AdaBound, self).__init__(params, defaults)

        self.base_lrs = list(map(lambda group: group['lr'], self.param_groups))

    def __setstate__(self, state):
        super(AdaBound, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsbound', False)

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group, base_lr in zip(self.param_groups, self.base_lrs):
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        'Adam does not support sparse gradients, please consider SparseAdam instead')
                amsbound = group['amsbound']

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    if amsbound:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if amsbound:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1

                if group['weight_decay'] != 0:
                    grad = grad.add(group['weight_decay'], p.data)

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                if amsbound:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = max_exp_avg_sq.sqrt().add_(group['eps'])
                else:
                    denom = exp_avg_sq.sqrt().add_(group['eps'])

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1

                # Applies bounds on actual learning rate
                # lr_scheduler cannot affect final_lr, this is a workaround to apply lr decay
                final_lr = group['final_lr'] * group['lr'] / base_lr
                lower_bound = final_lr * (1 - 1 / (group['gamma'] * state['step'] + 1))
                upper_bound = final_lr * (1 + 1 / (group['gamma'] * state['step']))
                step_size = torch.full_like(denom, step_size)
                step_size.div_(denom).clamp_(lower_bound, upper_bound).mul_(exp_avg)

                p.data.add_(-step_size)

        return loss

class AdaBoundW(torch.optim.Optimizer):
    """Implements AdaBound algorithm with Decoupled Weight Decay

    - **Paper**: [Adaptive Gradient Methods with Dynamic Bound of Learning Rate](https://arxiv.org/pdf/1902.09843.pdf).
    - **Implementation**: [GitHub Repo @ Luolc](https://github.com/Luolc/AdaBound)
        
    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): Adam learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        final_lr (float, optional): final (SGD) learning rate (default: 0.1)
        gamma (float, optional): convergence speed of the bound functions (default: 1e-3)
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsbound (boolean, optional): whether to use the AMSBound variant of this algorithm
    .. Adaptive Gradient Methods with Dynamic Bound of Learning Rate:
        https://openreview.net/forum?id=Bkg3g2R9FX
    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: float=(0.9, 0.999),
        final_lr: float=0.1,
        gamma: float=1e-3,
        eps: float=1e-8,
        weight_decay: float=0,
        amsbound: bool=False
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= final_lr:
            raise ValueError("Invalid final learning rate: {}".format(final_lr))
        if not 0.0 <= gamma < 1.0:
            raise ValueError("Invalid gamma parameter: {}".format(gamma))
        defaults = dict(lr=lr, betas=betas, final_lr=final_lr, gamma=gamma, eps=eps,
                        weight_decay=weight_decay, amsbound=amsbound)
        super(AdaBoundW, self).__init__(params, defaults)

        self.base_lrs = list(map(lambda group: group['lr'], self.param_groups))

    def __setstate__(self, state):
        super(AdaBoundW, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsbound', False)

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group, base_lr in zip(self.param_groups, self.base_lrs):
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        'Adam does not support sparse gradients, please consider SparseAdam instead')
                amsbound = group['amsbound']

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    if amsbound:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if amsbound:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                if amsbound:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = max_exp_avg_sq.sqrt().add_(group['eps'])
                else:
                    denom = exp_avg_sq.sqrt().add_(group['eps'])

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1

                # Applies bounds on actual learning rate
                # lr_scheduler cannot affect final_lr, this is a workaround to apply lr decay
                final_lr = group['final_lr'] * group['lr'] / base_lr
                lower_bound = final_lr * (1 - 1 / (group['gamma'] * state['step'] + 1))
                upper_bound = final_lr * (1 + 1 / (group['gamma'] * state['step']))
                step_size = torch.full_like(denom, step_size)
                step_size.div_(denom).clamp_(lower_bound, upper_bound).mul_(exp_avg)

                if group['weight_decay'] != 0:
                    decayed_weights = torch.mul(p.data, group['weight_decay'])
                    p.data.add_(-step_size)
                    p.data.sub_(decayed_weights)
                else:
                    p.data.add_(-step_size)

        return loss



import moai.nn.convolution.spherical as mis

import torch

__all__ = [
    "SphericalConv2d",
]

class SphericalConv2d(mis.SphericalPad2d):
    def __init__(self,
        in_channels: int,
        out_channels: int,
        kernel_size: int=3,
        dilation: int=1,
        stride: int=1,
        padding: int=0,
        groups: int=1,
        bias: bool=True
    ):
        super(SphericalConv2d, self).__init__(padding=padding if kernel_size > 1 else 0)
        self.conv2d = torch.nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            dilation=dilation,
            stride=stride,
            groups=groups,
            padding=0,
            padding_mode='zeros'
        )
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        #TODO: if kernel_size == 1 dont pad
        padded = super(SphericalConv2d, self).forward(x)
        return self.conv2d(padded)


from moai.validation.metrics.object_pose.rotation import _angular_error

import torch
import functools

__all__ = [
    "Accuracy2",
    "Accuracy5",
    "Accuracy10",
]

def _calculate_acc(
    gt_rot:     torch.Tensor,
    pr_rot:     torch.Tensor,
    gt_pos:     torch.Tensor,
    pr_pos:     torch.Tensor,
    threshold:  float
):
    """
        An estimated pose is correct if the average distance 
        is smaller than 5pixels.k◦,  k  cm as proposed  in  Shotton  et  al.  (2013).
        The  5◦,5cm metric considers an estimated pose to be correct if
        its rotation error is within 5◦ and the translation error is below 5cm.
        Provide also the results with 2◦, 2cm and 10◦, 10 cm.
        input: calculates the error per batch size
    """
    b, _, __  = gt_rot.shape
    rotation_error = _angular_error(gt_rot, pr_rot, radians=False)
    translation_error = torch.linalg.norm(gt_pos - pr_pos, ord=2, dim=-1) * 100.0 # in cm
    #count correct threshold
    condition_thres = torch.where(
        (rotation_error < threshold) & (translation_error < threshold),
        1.0, 0.0
    )
    count_thres = torch.count_nonzero(condition_thres)    
    acc = 100 * float(count_thres) / b
    return torch.tensor(acc).float()

class Accuracy(torch.nn.Module):
    def __init__(self,
        threshold: int=10,
    ):
        super(Accuracy, self).__init__()
        self.threshold = threshold

    def forward(self,
        pred_rotation:  torch.Tensor,
        gt_rotation:    torch.Tensor,
        pred_position:  torch.Tensor,
        gt_position:    torch.Tensor,
    ) -> torch.Tensor:
        return _calculate_acc(
            gt_rotation, pred_rotation,
            gt_position, pred_position,
            self.threshold
        )

Accuracy2 = functools.partial(
    Accuracy, 
    threshold=2
)

Accuracy5 = functools.partial(
    Accuracy, 
    threshold=5
)

Accuracy10 = functools.partial(
    Accuracy, 
    threshold=10
)


import torchvision
import typing
import torch
import logging

log = logging.getLogger(__name__)

__all__ = ['CIFAR10', 'CIFAR100']

#NOTE: normalization values taken from: https://github.com/kuangliu/pytorch-cifar/issues/19

__NORM_MEAN_STD__ = {
    'CIFAR10': {
        'mean': torch.Tensor([
            [[0.49139968]],     [[0.48215841]],     [[0.44653091]]
        ]),
        'std': torch.Tensor([
            [[0.24703223]],     [[0.24348513]],     [[0.26158784]]
        ]),
    },
    'CIFAR100': {
        'mean': torch.Tensor([
            [[0.50707516]],     [[0.48654887]],     [[0.44091784]]
        ]),
        'std': torch.Tensor([
            [[0.26733429]],     [[0.25643846]],     [[0.27615047]]
        ]),
    }
}

def _identity(x: torch.Tensor) -> torch.Tensor:
    return x

class CIFAR10(torchvision.datasets.CIFAR10):
    def __init__(self,
        root_path:      str,
        split:          str,
        download:       bool=True,
    ):
        super(CIFAR10, self).__init__(
            root=root_path,
            train=split == 'train',
            download=download,
            transform=torchvision.transforms.Compose([
                # torchvision.transforms.RandomCrop(size=32, padding=4),
                torchvision.transforms.RandomHorizontalFlip() if split == 'train'\
                    else torchvision.transforms.Lambda(_identity),
                torchvision.transforms.ToTensor(),
                torchvision.transforms.Normalize(
                    mean=__NORM_MEAN_STD__[__class__.__name__]['mean'].squeeze().tolist(),
                    std=__NORM_MEAN_STD__[__class__.__name__]['std'].squeeze().tolist(),
                    inplace=True,
                )
            ])
        )
        log.info(f"Loaded {len(self)} items from CIFAR10")

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        image, label = super(CIFAR10, self).__getitem__(index)
        return {
            'color': image,
            'label': label,
            'dataset_mean': __NORM_MEAN_STD__[__class__.__name__]['mean'],
            'dataset_std': __NORM_MEAN_STD__[__class__.__name__]['std'],
        }

class CIFAR100(torchvision.datasets.CIFAR100):
    def __init__(self,
        root_path:      str,
        split:          str,
        download:       bool=True,
    ):
        super(CIFAR100, self).__init__(
            root=root_path,
            train=split == 'train',
            download=download,
            transform=torchvision.transforms.Compose([
                # torchvision.transforms.RandomCrop(size=32, padding=4),
                torchvision.transforms.RandomHorizontalFlip() if split == 'train'\
                    else torchvision.transforms.Lambda(_identity),
                torchvision.transforms.ToTensor(),                
                torchvision.transforms.Normalize(
                    mean=__NORM_MEAN_STD__[__class__.__name__]['mean'].squeeze().tolist(),
                    std=__NORM_MEAN_STD__[__class__.__name__]['std'].squeeze().tolist(),
                    inplace=True,
                )
            ])
        )
        log.info(f"Loaded {len(self)} items from CIFAR100")

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        image, label = super(CIFAR100, self).__getitem__(index)
        return {
            'color': image,
            'label': label,
            'dataset_mean': __NORM_MEAN_STD__[__class__.__name__]['mean'],
            'dataset_std': __NORM_MEAN_STD__[__class__.__name__]['std'],
        }


import moai.nn.convolution as miconv
import moai.nn.sampling.spatial.upsample as miup

import torch
import typing
import omegaconf.omegaconf

__all__ = ["Convolutional"]

class Convolutional(torch.nn.Module):
    def __init__(self,
        configuration:      omegaconf.DictConfig,
        convolution:        omegaconf.DictConfig,
        prediction:         omegaconf.DictConfig,        
        upscale:            omegaconf.DictConfig, 
    ):
        super(Convolutional, self).__init__()
        module_list = torch.nn.ModuleList()
        in_features, out_features = configuration.bottleneck_features, configuration.out_features
        next_features = in_features // configuration.feature_block_divisor
        for block_size in configuration.blocks:
            module_list.append(miup.make_upsample(
                upscale_type=upscale.type,
                features=in_features,
                **upscale.params
            ))
            for b in range(0, block_size):
                module_list.append(miconv.make_conv_block(
                        block_type="conv2d",
                        convolution_type=convolution.type,
                        activation_type=convolution.activation.type,
                        in_features=in_features if b == 0 else next_features,
                        out_features=next_features,
                        convolution_params=convolution.params,
                        activation_params=convolution.activation.params
                    )
                )
            in_features = next_features
            next_features //= configuration.feature_block_divisor
        module_list.append(miconv.make_conv_block(
                block_type="conv2d",                      #TODO: check if possible to modify this
                convolution_type=prediction.convolution.type,  #TODO: check if this type can resolve sconv2d blocks
                activation_type=prediction.activation.type,
                in_features=in_features,
                out_features=out_features,
                convolution_params=prediction.convolution.params,
                activation_params=prediction.activation.params
            )
        )
        self.sequential = torch.nn.Sequential(*module_list)

    def forward(self, 
        features: torch.Tensor
    )-> torch.Tensor:
        return self.sequential(features)


import torch
import typing

__all__ = ['AdamT']

#NOTE: from https://github.com/xuebin-zh/AdamT
#NOTE: described in https://arxiv.org/pdf/2001.06130.pdf

class AdamT(torch.optim.Optimizer):
    """Implements the AdamT algorithm.
    
    - **Paper**: [ADAMT: A Stochastic Optimization with Trend Correction Scheme](https://arxiv.org/pdf/2001.06130.pdf).
    - **Implementation**: [GitHub @ xuebin-zh](https://github.com/xuebin-zh/AdamT).

    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: typing.Tuple[float, float, float, float]=(0.9, 0.999, 0.9, 0.999),
        phis: typing.Tuple[float, float]=(1, 1),
        eps: float=1e-8,
        weight_decay: float=0
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= betas[2] < 1.0:
            raise ValueError("Invalid beta parameter at index 2: {}".format(betas[2]))
        if not 0.0 <= betas[3] < 1.0:
            raise ValueError("Invalid beta parameter at index 3: {}".format(betas[3]))
        if not 0.0 <= phis[0] <= 100.0:
            raise ValueError("Invalid phi parameter at index 0: {}".format(phis[0]))
        if not 0.0 <= phis[1] <= 100.0:
            raise ValueError("Invalid phi parameter at index 1: {}".format(phis[1]))
        defaults = dict(lr=lr, betas=betas, phis=phis, eps=eps, weight_decay=weight_decay)
        super(AdamT, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(AdamT, self).__setstate__(state)

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("Adam does not support sparse gradients, please consider SparseAdam instead")

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    # First-order
                    # state["m"] = grad
                    state["m"] = torch.zeros_like(p.data)
                    # First-order level information
                    # state["L_m"] = grad
                    state["L_m"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    # Second-order
                    # state["v"] = grad * grad
                    state["v"] = torch.zeros_like(p.data)
                    # Second-order level information
                    # state["L_v"] = grad * grad
                    state["L_v"] = torch.zeros_like(p.data)
                    # Holt's linear trend information for first-order
                    state["B_m"] = torch.zeros_like(p.data)
                    # Holt's linear trend information for second-order
                    state["B_v"] = torch.zeros_like(p.data)

                m, v = state["m"], state["v"]
                L_m, L_v, B_m, B_v = state["L_m"], state["L_v"], state["B_m"], state["B_v"]
                beta1, beta2, beta3, beta4 = group["betas"]
                phi1, phi2 = group["phis"]

                state["step"] += 1

                if group["weight_decay"] != 0:
                    grad.add_(group["weight_decay"], p.data)

                # Partially update first-order trend information
                B_m.mul_(beta3 * phi1).add_((beta3 - 1), L_m)
                # Partially update second-order trend information
                B_v.mul_(beta4 * phi2).add_((beta4 - 1), L_v)
                # Fully update first-order level information
                torch.add(beta1 * m, (1 - beta1) * grad, out=L_m)
                # Fully update second-order level information
                torch.addcmul(beta2 * v, (1 - beta2), grad, grad, out=L_v)
                # Fully update first-order trend information
                B_m.add_((1 - beta3), L_m)
                # Fully update second-order trend information
                B_v.add_((1 - beta4), L_v)
                # Update first-order and second-order
                torch.add(L_m, phi1 * B_m, out=m)
                torch.add(L_v, phi2 * B_v, out=v)
                # Bias correction
                # m_hat = torch.div(L_m, (1 - beta1 ** state["step"])) + \
                #         torch.div(phi1 * B_m, (1 - beta3 ** state["step"]))
                # v_hat = torch.div(L_v, (1 - beta2 ** state["step"])) + \
                #         torch.div(phi2 * B_v, (1 - beta4 ** state["step"]))
                m_hat = torch.div(L_m, (1 - beta1 ** state["step"])) + \
                        torch.div(B_m, ((1 - beta3) * torch.div(1 - (beta3 * phi1) ** state["step"], 1 - beta3 * phi1)))

                v_hat = torch.div(L_v, (1 - beta2 ** state["step"])) + \
                        torch.div(B_v, ((1 - beta4) * torch.div(1 - (beta4 * phi2) ** state["step"], 1 - beta4 * phi2)))
                # Update the parameters

                # if len(torch.nonzero(torch.isnan(torch.div(m_hat, (v_hat.sign() * torch.sqrt(v_hat.abs()) + group["eps"]))))) != 0:
                    # print(torch.div(m_hat, (v_hat.sign() * torch.sqrt(v_hat.abs()) + group["eps"])))
                # print(v_hat.sign() * torch.sqrt(v_hat.abs()))
                # print(torch.div(m_hat, (v_hat.sign() * torch.sqrt(v_hat.abs()) + group["eps"])))
                # print(torch.div(m_hat, (v_hat.sign() * torch.sqrt(v_hat.abs()) + group["eps"])))
                # p.data.add_(-group["lr"], torch.div(m_hat, (v_hat.sign() * torch.sqrt(v_hat.abs()) + group["eps"])))
                p.data.add_(-group["lr"], torch.div(m_hat, (torch.sqrt(v_hat.abs()) + group["eps"])))
                # print(v_hat.sign())
                # p.data.add_(-group["lr"], torch.div(m_hat, (v_hat.sign() * torch.sqrt(v_hat.abs() + group["eps"]))))

        return loss


from moai.parameters.initialization.schemes.kaiming import Kaiming
from moai.parameters.initialization.schemes.xavier import Xavier
from moai.parameters.initialization.schemes.prediction_bias import PredictionBias
from moai.parameters.initialization.schemes.zero_params import ZeroParams

__all__ = [    
    "Kaiming",
    "PredictionBias",
    "Xavier",
    "ZeroParams",
]


from moai.utils.arguments import ensure_numeric_list
from moai.monads.utils import expand_spatial_dims

import torch
import typing

__all__ = ["Grid"]

def _create_inclusive(size: typing.List[int]) -> torch.Tensor:
    return torch.stack(
        torch.meshgrid(
            *[torch.linspace(0.0, 1.0 - (1.0 / dim), dim) for dim in size],
        )
    ).unsqueeze(0)

def _create_exclusive(size: typing.List[int]) -> torch.Tensor:    
    return torch.stack(
        torch.meshgrid(
            *[(torch.arange(dim) + 0.5) / dim for dim in size],
        )
    ).unsqueeze(0)

def _scale_to_size(grid: torch.Tensor, size: typing.List[int]) -> torch.Tensor:
    sizes = torch.Tensor([*size]).unsqueeze(0)
    sizes = expand_spatial_dims(sizes, grid)
    return grid * sizes

__MODES__ = ['ndc', 'coord', 'norm']

__MODE_CONVERTERS__ = {
    'ndc': lambda g, _: torch.addcmul(torch.scalar_tensor(-1.0), g, torch.scalar_tensor(2.0)), # from [0, 1] to [-1, 1]
    'coord': _scale_to_size, # from [0, 1] to [0, S] where S := { W, (H), (D) }
    'norm': lambda g, _: g, # passthrough, remains at [0, 1]
}

class Grid(torch.nn.Module):
    __XYZ__ = 'zyx'

    def __init__(self,
        mode:           str='ndc', # 'ndc', 'coords', 'norm'
        width:          int=256,
        height:         int=1,
        depth:          int=1,
        inclusive:      bool=True,
        order:          str='xyz', # order of coords
        persistent:     bool=True, # save and re-use the grid buffer
    ):
        super(Grid, self).__init__()
        size = [n for n in [depth, height, width] if n > 1]
        unit_grid = _create_inclusive(size) if inclusive\
            else _create_exclusive(size)
        grid = __MODE_CONVERTERS__[mode](unit_grid, size)
        order = order[:len(size)]
        size_diff = len(Grid.__XYZ__) - len(order)
        indices = [Grid.__XYZ__.index(c) - size_diff for c in order if c in Grid.__XYZ__]
        grid = torch.stack([
            grid[:, i, ...] for i in indices
        ], dim=1)
        self.register_buffer("grid", grid, persistent)

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        b = tensor.shape[0]
        return self.grid.expand(b, *self.grid.shape[1:])

class MultiScaleGrid(torch.nn.Module):
    def __init__(self,
        mode:           str='ndc', # 'ndc', 'coords', 'norm'
        width:          typing.Sequence[int]=256,
        height:         typing.Sequence[int]=1,
        depth:          typing.Sequence[int]=1,
        inclusive:      bool=True,
        reverse:        bool=False,
    ):
        super(MultiScaleGrid, self).__init__()
        sizes = []
        for w, h, d in zip(
            ensure_numeric_list(width),
            ensure_numeric_list(height),
            ensure_numeric_list(depth)
        ):
            sizes += [n for n in [d, h, w] if n > 1]
        for i, size in enumerate(sizes):
            unit_grid = _create_inclusive(size) if inclusive\
                else _create_exclusive(size)
            grid = __MODE_CONVERTERS__[mode](unit_grid, size)
            grid = torch.flip(grid, dims=[1]) if not reverse else grid
            self.register_buffer("grid_" + str(i), grid)

    def forward(self, tensor: torch.Tensor) -> typing.List[torch.Tensor]:
        b = tensor.shape[0]
        grids = []
        i = 0
        while hasattr("grid_" + str(i)):
            grid = getattr("grid_" + str(i))
            grids.append(grid.expand(b, *grid.shape[1:]))
            i += 1
        return grids


import torch
import math
import typing

class SAdam(torch.optim.Optimizer):
    """Implements SAdam/SAmsgrad algorithm.

    - **Paper**: [Calibrating the Learning Rate for Adaptive Gradient Methods to Improve Generalization Performance](https://arxiv.org/pdf/1908.00700v1.pdf)
    - **Implementation**: [GitHub @ neilliang90/Sadam](https://github.com/neilliang90/Sadam/)

	Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, required): learning rate (default: 1e-1)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        partial (float, optional): partially adaptive parameter
    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-1,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-8,
        weight_decay: float=0,
        amsgrad: bool=True,
        partial: float = 1/4,
        transformer: str='softplus',
        grad_transf: str='square',
        smooth: float=50,
        hist: bool=False
    ):
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, amsgrad=amsgrad, partial = partial, transformer=transformer,  hist=hist, grad_transf = grad_transf , smooth = smooth)
        super(SAdam, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()
        denom_list = []
        denom_inv_list = []
        m_v_eta = []
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')
                amsgrad = group['amsgrad']
                partial = group['partial']
                grad_transf = group['grad_transf']
                smooth = group['smooth']

                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    if amsgrad:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if amsgrad:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1
                if group['weight_decay'] != 0:
                    grad = grad.add(group['weight_decay'], p.data)
                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                ################ tranformer for grad ###########################################################    
                if grad_transf == 'square':
                    grad_tmp = grad**2
                elif grad_transf == 'abs':
                    grad_tmp = grad.abs()
                    
                exp_avg_sq.mul_(beta2).add_((1 - beta2)*grad_tmp)
                    
                if amsgrad:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = max_exp_avg_sq.clone()
                else:
                    denom = exp_avg_sq.clone()
                if grad_transf == 'square':
                    denom.sqrt_() 
 
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1
                ################ tranformer for denom ###########################################################        
                if  group['transformer'] =='softplus':
                    sp = torch.nn.Softplus( smooth)
                    denom = sp( denom)
                    p.data.addcdiv_(-step_size, exp_avg, denom )                    
                else:
                    p.data.addcdiv_(-step_size, exp_avg, (denom.add_(group['eps']))**(partial*2))
                ################ logger ###########################################################       
                if group['hist']:    
                    if  group['transformer'] =='softplus':
                        denom_list.append(denom.reshape( 1, -1).squeeze())
                        denom_inv_list.append( (1/denom).reshape( 1, -1).squeeze() )
                        m_v_eta.append( (-step_size*torch.mul(exp_avg, (1/denom ))).reshape( 1, -1).squeeze()  )
                    else:
                        denom_list.append(denom.reshape( 1, -1).squeeze())
                        denom_inv_list.append( (1/(denom)**(partial*2)).reshape( 1, -1).squeeze() ) 
                        m_v_eta.append( (-step_size*torch.mul(exp_avg, (1/(denom)**(partial*2) ))).reshape( 1, -1).squeeze()  )
                            
        return {
            "denom": denom_list,
            "denom_inv": denom_inv_list,
            "m_v_eta":m_v_eta 
        }

    


from moai import __version__ as miV
from moai.utils.engine import NoOp as NoInterval
from moai.parameters.optimization import NoOp as NoOptimization
from moai.parameters.optimization.schedule import NoOp as NoScheduling
from moai.parameters.initialization import Default as NoInit
from moai.validation import (
    NoOp as NoValidation,
    Collection as DefaultValidation,
)
from moai.supervision import (
    NoOp as NoSupervision,
    Weighted as DefaultSupervision,
)
from moai.data.iterator import Indexed
from moai.parameters.optimization.optimizers import LBFGS as miLBFGS

import moai.utils.parsing.rtp as mirtp

import torch
import pytorch_lightning
import omegaconf.omegaconf
import hydra.utils as hyu
import typing
import toolz
import logging
import inspect
import functools

log = logging.getLogger(__name__)

__all__ = ['Optimizer']

def _create_supervision_block(
    cfg: omegaconf.DictConfig,
    force: bool=True
):
    if force and not cfg:
        log.warning("Empty supervision block in feedforward model.")
    if not cfg:
        return NoSupervision()
    if '_target_' not in cfg:
        return DefaultSupervision(**cfg)
    else:
        return hyu.instantiate(cfg)

def _create_validation_block(
    cfg: omegaconf.DictConfig,
    force: bool=True
):
    if force and not cfg:
        log.warning("Empty validation block in feedforward model.")    
    if not cfg:
        return NoValidation()
    if '_target_' not in cfg:
        return DefaultValidation(**cfg)
    else:
        return hyu.instantiate(cfg)

def _create_processing_block(
    cfg: omegaconf.DictConfig, 
    attribute: str, 
    monads: omegaconf.DictConfig,
    force: bool=False
):
    if force and not cfg and attribute in cfg:
        log.warning(f"Empty processing block ({attribute}) in feedforward model.")
    return hyu.instantiate(getattr(cfg, attribute), monads)\
        if cfg and attribute in cfg else torch.nn.Identity()

def _create_interval_block(
    cfg: omegaconf.DictConfig,
    force: bool=False
):
    if force and not cfg:
        log.warning("Empty interval block in feedforward model.")
    return hyu.instantiate(cfg) if cfg else NoInterval()

def _create_optimization_block(
    cfg: omegaconf.DictConfig,
    params: typing.Union[typing.Iterable[torch.Tensor], typing.Dict[str, torch.Tensor]],
    force: bool=False
):
    if force and not cfg:
        log.warning("No optimizer in feedforward model.")
    return hyu.instantiate(
        # cfg,
        #TODO: hack to get resolved values, should update when omegaconf 2.1 is out and explicit resolve is available
        omegaconf.OmegaConf.create(omegaconf.OmegaConf.to_container(cfg, resolve=True)),
        params
    ) if cfg else NoOptimization(params)

def _create_scheduling_block(
    cfg: omegaconf.DictConfig, 
    optimizers: typing.Sequence[torch.optim.Optimizer],
    force: bool=False
):
    if force and not cfg:
        log.warning("No scheduling used in feedforward model.")
    return hyu.instantiate(cfg, optimizers) if cfg else NoScheduling(optimizers)

def _assign_data(cfg: omegaconf.DictConfig) -> omegaconf.DictConfig:
    has_train = hasattr(cfg, 'train')
    has_test = hasattr(cfg, 'test')
    has_val = hasattr(cfg, 'val')
    if not has_train and not has_test and not has_val:
        log.error("No data have been included in the configuration. Please add the necessary \'- data/[split]/dataset/*: *\' entries.")
    if not has_test and has_val:
        test = omegaconf.OmegaConf.create({'test': cfg.val})
        omegaconf.OmegaConf.set_struct(cfg, False)
        cfg.merge_with(test)
        omegaconf.OmegaConf.set_struct(cfg, True)
        log.warning("No test dataset has been defined, using the validation dataset as a test dataset.")
    if not has_val and has_test:
        val = omegaconf.OmegaConf.create({'val': cfg.test})
        omegaconf.OmegaConf.set_struct(cfg, False)
        cfg.merge_with(val)
        omegaconf.OmegaConf.set_struct(cfg, True)
        log.warning("No validation dataset has been defined, using the test dataset as a validation dataset.")
    return cfg

class PerBatch(torch.nn.Identity, pytorch_lightning.Callback):
    def __init__(self):
        super(PerBatch, self).__init__()

    def on_train_batch_start(self,
        trainer: pytorch_lightning.Trainer,
        pl_module: pytorch_lightning.LightningModule,
        batch: typing.Dict[str, typing.Union[torch.Tensor, typing.Sequence[torch.Tensor], typing.Dict[str, torch.Tensor]]],
        batch_idx: int,
        unused: typing.Optional[int] = 0,
    ) -> None:
        """Called when the train batch begins."""
        pl_module.initialize_parameters()
        if batch_idx > 0:
            trainer.accelerator.setup_optimizers(trainer)        
        if 'inference' in pl_module.mode:
            with torch.no_grad():
                pl_module.preprocess(batch)
                pl_module(batch)
                pl_module.initialize(batch)
            if pl_module.optimized_predictions:
                for key, values in pl_module.optimized_predictions.items():
                    for optim in filter(lambda o: o.name == key, trainer.accelerator.optimizers):
                        for v in values:
                            params = pl_module.predictions[v]
                            if isinstance(optim, torch.optim.LBFGS) or\
                                isinstance(optim, miLBFGS):
                                    optim._params.append(params)
                            else:
                                optim.param_groups.append({'params': params})
        pl_module.optimization_step = 0

    def on_train_batch_end(
        self,
        trainer: pytorch_lightning.Trainer,
        pl_module: pytorch_lightning.LightningModule,
        outputs: typing.Dict[str, typing.Union[torch.Tensor, typing.Sequence[torch.Tensor], typing.Dict[str, torch.Tensor]]],
        batch: typing.Dict[str, typing.Union[torch.Tensor, typing.Sequence[torch.Tensor], typing.Dict[str, torch.Tensor]]],
        batch_idx: int,
        unused: typing.Optional[int] = 0,
    ) -> None:
        """Called when the train batch ends."""        
        metrics = pl_module.validation(batch)
        pl_module.log_dict(metrics, prog_bar=True, logger=False, on_epoch=False, on_step=True, sync_dist=True)
        log_metrics = toolz.keymap(lambda k: f"val_{k}", metrics)
        pl_module.log_dict(log_metrics, prog_bar=False, logger=True, on_epoch=False, on_step=True, sync_dist=True)        
        pl_module.visualizer(batch)
        pl_module.exporter(batch)

from moai.monads.execution.cascade import _create_accessor

#TODO: vary w.r.t mode infer/predict to change clone/copy behaviour
def _create_assigner(key: str) -> typing.Callable[[torch.nn.Module, torch.Tensor], None]:
    key_list = [key] if isinstance(key, str) else key
    splits = [k.split('.') for k in key_list]
    def _assign(m: torch.nn.Module, t: torch.Tensor, keys: typing.Sequence[str]):
        for split in keys:
            to_set = toolz.reduce(getattr, split, m)
            to_set.copy_(t.clone())
            # to_set.copy_(t)    
    return functools.partial(_assign, keys=splits)

class Optimizer(pytorch_lightning.LightningModule):
    def __init__(self, 
        inner:              omegaconf.DictConfig,
        configuration:      omegaconf.DictConfig,
        data:               omegaconf.DictConfig=None,
        parameters:         omegaconf.DictConfig=None,
        feedforward:        omegaconf.DictConfig=None,
        monads:             omegaconf.DictConfig=None,        
        supervision:        omegaconf.DictConfig=None,
        validation:         omegaconf.DictConfig=None,        
        visualization:      omegaconf.DictConfig=None,
        export:             omegaconf.DictConfig=None,
        hyperparameters:    typing.Union[omegaconf.DictConfig, typing.Mapping[str, typing.Any]]=None,
    ):
        super(Optimizer, self).__init__()
        self.model = hyu.instantiate(inner)
        self.fwds = []
        params = inspect.signature(self.model.forward).parameters
        model_in = list(zip(*[mirtp.force_list(configuration.io[prop]) for prop in params]))
        model_out = mirtp.split_as(mirtp.resolve_io_config(configuration.io['out']), model_in)
        self.res_fill = [mirtp.get_result_fillers(self.model, out) for out in model_out]
        get_filler = iter(self.res_fill)
        self.predictions = { }
        self.setters = []
        for keys in model_in:          
            accessors = [_create_accessor(k if isinstance(k, str) else k[0]) for k in keys]  
            self.fwds.append(lambda td,
                tk=keys,
                acc=accessors,
                args=params.keys(),
                model=self.model,
                filler=next(get_filler):
                    filler(td, model(**dict(zip(args,
                        list(acc[i](td) if type(k) is str
                            else list(td[j] for j in k)
                        for i, k in enumerate(tk))
                    )))) # list(td[k] if type(k) is str
            )
            self.setters.extend(toolz.concat(model_out))
        # self.mode = omegaconf.OmegaConf.select(configuration, 'mode', default='inference') #NOTE: update when omegaconf/hydra updates
        self.mode = omegaconf.OmegaConf.select(configuration, 'mode')
        if self.mode == 'inference':
            self.model = self.model.eval()
            self.model.requires_grad_(False)
        self.data = _assign_data(data)
        self.initializers = [
            (i, _create_assigner(o)) for i, o in configuration.initialize.items()
        ] if configuration.initialize is not None else []
        self.assigners = [
            (i, _create_assigner(o)) for i, o in toolz.dissoc(configuration.assign, 'parameters').items()
        ] if configuration.assign is not None else []
        self.assigned_params = toolz.get_in(['assign', 'parameters'], configuration)
        self.assigned_params = [self.assigned_params] if isinstance(self.assigned_params, str) else self.assigned_params
        self.optimized_predictions = configuration.optimize_predictions
        self.initializer = parameters.initialization if parameters is not None else None
        self.preprocess = _create_processing_block(feedforward, "preprocess", monads=monads)
        self.postprocess = _create_processing_block(feedforward, "postprocess", monads=monads)
        self.optimizer_configs = toolz.get_in(['optimization', 'optimizers'], parameters) or []
        self.parameter_selectors = parameters.selectors or []
        self.scheduler_configs = toolz.get_in(['optimization', 'schedulers'], parameters) or { }
        self.supervision = torch.nn.ModuleDict()
        self.params_optimizers = []
        self.stages = []
        optimization_process = toolz.get_in(['optimization', 'process'], parameters) or { }
        log.info(f"Optimization with {len(optimization_process)} stages.")
        for stage, cfg in optimization_process.items():
            self.stages.append(stage)
            log.info(
                f"Setting up the '{stage}' stage using the " + str(cfg.optimizer or "same") + " optimizer, "
                f"optimizing the " + str(cfg.selectors or "same") + " parameters"
            )
            optimizer = cfg.optimizer
            iterations = cfg.iterations
            scheduler = cfg.scheduler
            selector = cfg.selectors
            self.params_optimizers.append((optimizer, stage, iterations, selector, scheduler))
            objective = cfg.objective #TODO: merge confs from supervision and objectives
            self.supervision[stage] = _create_supervision_block(
                omegaconf.OmegaConf.merge(supervision, objective)
            )
        self.validation = _create_validation_block(validation) #TODO: change this, "empty processing block" is confusing        
        self.visualizer = _create_interval_block(visualization)
        self.exporter = _create_interval_block(export)        
        #NOTE: __NEEDED__ for loading checkpoint?
        hparams = hyperparameters if hyperparameters is not None else { }
        hparams.update({'moai_version': miV})
        #NOTE: @PTL1.5 self.hparams =  hparams
        self.hparams.update(hparams)
        self.optimization_step = 0
        self.per_batch = PerBatch()
        self.prediction_stages = []

    def initialize_parameters(self) -> None:
        init = hyu.instantiate(self.initializer) if self.initializer else NoInit()
        init(self)

    def initialize(self,
        tensors: typing.Dict[str, torch.Tensor]
    ) -> None:
        for i, a in self.initializers:
            a(self, tensors[i])

    def assign(self,
        tensors: typing.Dict[str, torch.Tensor]
    ) -> None:
        for i, a in self.assigners:
            a(self, tensors[i])

    def forward(self,
        tensors: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        for f in self.fwds:
            f(tensors)
        for k in self.setters:
            self.predictions[k] = tensors[k]
        return tensors

    def training_step(self, 
        batch:                  typing.Dict[str, torch.Tensor],
        batch_idx:              int,
        optimizer_idx:          int=0,
    ) -> typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]:
        batch['__moai__'] = { 'batch_index': batch_idx }
        batch['__moai__']['optimization_stage'] = self.stages[optimizer_idx]
        td = self.preprocess(batch)
        if 'predict' in self.mode and\
            self.stages[optimizer_idx] in self.prediction_stages: #TODO: only run in needed stages
                td = self(td)
        td = self.postprocess(td)
        total_loss, losses = self.supervision[self.stages[optimizer_idx]](td)
        #TODO: should add loss maps as return type to be able to forward them for visualization
        losses = toolz.keymap(lambda k: f"train_{k}", losses)
        losses.update({'total_loss': total_loss})
        self.log_dict(losses, prog_bar=False, logger=True)
        self.optimization_step += 1
        return { 'loss': total_loss, 'tensors': td }

    def training_step_end(self, 
        train_outputs: typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]
    ) -> None:
        train_outputs['tensors']['__moai__']['optimization_step'] = self.optimization_step
        if (self.optimization_step + 1) and (self.optimization_step % self.visualizer.interval == 0):
            self.visualizer(train_outputs['tensors'])
        if (self.optimization_step + 1) and (self.optimization_step % self.exporter.interval == 0):
            self.exporter(train_outputs['tensors'])
        return train_outputs['loss']

    def configure_optimizers(self) -> typing.Tuple[typing.List[torch.optim.Optimizer], typing.List[torch.optim.lr_scheduler._LRScheduler]]:
        log.info(f"Configuring optimizers and schedulers")                
        optimizers, schedulers = [], []        
        for optimizer, name, iterations, params, schedule in self.params_optimizers:
            if optimizer is None and params is None:
                optimizers.append(optimizers[-1])
                schedulers.append(schedulers[-1])
                getattr(optimizers[-1], 'iterations').append(iterations)
                getattr(optimizers[-1], 'name').append(name)
            elif isinstance(optimizer, str):
                parameters = hyu.instantiate(self.parameter_selectors[params])(self) if isinstance(params, str) else\
                    list(toolz.concat(hyu.instantiate(self.parameter_selectors[p])(self) for p in params))
                #TODO: parameter construction is very slow
                optimizers.append(_create_optimization_block(
                    self.optimizer_configs[optimizer], parameters
                ))#TODO: check if it works with a list of parameters
                setattr(optimizers[-1], 'iterations', [iterations])
                setattr(optimizers[-1], 'name', [name])
                schedulers.append(_create_scheduling_block(
                    self.scheduler_configs.get(schedule, None), [optimizers[-1]]
                ))
                if any(p in (self.assigned_params or []) for p in params):
                    optimizers[-1].assign = self.assign
                    self.prediction_stages.append(name)
            else:
                parameters = [
                    hyu.instantiate(self.parameter_selectors[par])(self) if isinstance(par, str)
                    else list(toolz.concat(hyu.instantiate(self.parameter_selectors[p])(self) for p in par))
                    for par in params
                ]
                alternated = [_create_optimization_block(
                        self.optimizer_configs[o], param  
                    ) for o, param in zip(optimizer, parameters)
                ]
                for ap in self.assigned_params or []:
                    alternated[params.index(ap)].assign = self.assign
                    self.prediction_stages.append(name)
                optimizers.append(AlternatingOptimizer(
                    alternated, toolz.concat(parameters))                    
                )
                setattr(optimizers[-1], 'iterations', [iterations])
                setattr(optimizers[-1], 'name', [name])                
                schedulers.append(_create_scheduling_block(
                    self.scheduler_configs.get(schedule, None), [optimizers[-1]]
                ))
        return (
            optimizers,
            list(map(lambda s: s.schedulers[0] if isinstance(s, NoScheduling) else s, schedulers))
        )

    def optimizer_step(
        self,
        epoch: int,
        batch_idx: int,
        optimizer: typing.Union[torch.optim.Optimizer, pytorch_lightning.core.optimizer.LightningOptimizer],
        optimizer_idx: int = 0,
        optimizer_closure: typing.Optional[typing.Callable[[], typing.Any]] = None,
        on_tpu: bool = False,
        using_native_amp: bool = False,
        using_lbfgs: bool = False,
    ) -> None:        
        optimizer.step(closure=optimizer_closure)

    def train_dataloader(self) -> torch.utils.data.DataLoader:
        #NOTE: we are in train mode as we may need to optimize weights,
        # but semantically its test time so we use the test set
        if hasattr(self.data.test.iterator, '_target_'):
            log.info(f"Instantiating ({self.data.test.iterator._target_.split('.')[-1]}) test set data iterator")
            test_iterator = hyu.instantiate(self.data.test.iterator)
        else:
            test_iterator = Indexed(
                self.data.test.iterator.datasets,
                self.data.test.iterator.augmentation if hasattr(self.data.test.iterator, 'augmentation') else None,
            )
        if not hasattr(self.data.test, 'loader'):
            log.error("Test data loader missing. Please add a data loader (i.e. \'- data/test/loader: torch\') entry in the configuration.")
        else:
            test_loader = hyu.instantiate(self.data.test.loader, test_iterator, 
                shuffle=False, batch_size=1
            )
        return test_loader

class AlternatingOptimizer(torch.optim.Optimizer):
    def __init__(self, 
        optimizers: typing.Iterable[torch.optim.Optimizer],
        parameters: typing.Iterable[torch.nn.parameter.Parameter],
    ):
        self.optimizers = optimizers
        super(AlternatingOptimizer, self).__init__(
            parameters, {'lr': 1.0}
        )

    def step(self, closure):
        for o in self.optimizers:
            o.step(closure)
            if hasattr(o, 'assign'):
                with torch.no_grad():
                    o.assign(closure.args[3]._step_fn.args[0])


import torch
import typing

class NoOp(torch.nn.ModuleDict):
    def __init__(self):
        super(NoOp, self).__init__()

    def forward(self, 
        tensors: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        return torch.scalar_tensor(0.0), { }


import torch
import logging

log = logging.getLogger(__name__)

#NOTE: see this https://medium.com/@me_26124/debugging-neural-networks-6fa65742efd

__all__ = ['AnomalyDetection']

class AnomalyDetection(object):
    def __init__(self, 

    ):
        log.info(f"Toggling PyTorch anomaly detection ON.")
        # torch.set_anomaly_enabled()
        torch.autograd.set_detect_anomaly(True)


from moai.parameters.initialization.default import Default
from moai.parameters.initialization.single import Initializer as Single
from moai.parameters.initialization.cascade import Initializers as Cascade
from moai.parameters.initialization.named import Named

__all__ = [
    "Default",
    "Single", 
    "Cascade",
    "Named",
]


from moai.monads.sampling.torch import (
    Interpolate,
    BilinearDownsample_x2,
    NearestDownsample_x2,
    BilinearSampling,
    NearestSampling,
)

__all__ = [
    "Interpolate",
    "BilinearDownsample_x2",
    "NearestDownsample_x2",
    "BilinearSampling",
    "NearestSampling",
]


from moai.monads.math.clamp import Clamp
from moai.monads.math.common import (
    Abs,
    Scale,
    Plus,
    Minus,
    Multiply,
    PlusOne,
    MinusOne,
)
from moai.monads.math.normalization import (
    Znorm,
    MinMaxNorm,
)

__all__ = [
    "Abs",
    "Scale",
    "Plus",
    "Minus",
    "Multiply",
    "PlusOne",
    "MinusOne",
    "Clamp",  
    "Znorm",
    "MinMaxNorm",
]


from moai.validation.collection import Metrics
from moai.utils.arguments import ensure_string_list

import torch
import omegaconf.omegaconf
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ['Indicators']

class Indicators(Metrics):
    execs: typing.List[typing.Callable] = []

    def __init__(self, 
        metrics:            omegaconf.DictConfig={},
        indicators:         typing.Sequence[str]=[],
        **kwargs:           typing.Mapping[str, typing.Any]
    ):
        super(Indicators, self).__init__(metrics=metrics, **kwargs)
        self.expressions = { }
        for name, expression in indicators.items():
            self.expressions[name] = expression\
                .replace('[', 'returned[\'').replace(']', '\']')

    def forward(self,
        tensors: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        metrics = { }
        for exe in self.execs:
            exe(tensors, metrics)
        returned = { }
        for k, m in metrics.items():
            returned[f'{k}'] = m
        for name, expression in self.expressions.items():
            returned[name] = eval(expression)
        return returned


import torch
import typing

__all__ = ['AggMo']

T = typing.TypeVar('T', bound='AggMo')

#NOTE: from https://github.com/jettify/pytorch-optimizer/blob/master/torch_optimizer/aggmo.py

class AggMo(torch.optim.Optimizer):
    r"""Implements Aggregated Momentum Gradient Descent.
    
    - **Paper**: [Aggregated Momentum: Stability Through Passive Damping](https://arxiv.org/pdf/1804.00325.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)    

    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.AggMo(model.parameters(), lr=0.1)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

     __ https://arxiv.org/abs/1804.00325

    Note:
        Reference code: https://github.com/AtheMathmo/AggMo/blob/master/aggmo.py  # noqa
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: typing.Union[typing.List[float], typing.Tuple[float, ...]]=(0.0, 0.9, 0.99),
        weight_decay: float=0,
    ) -> None:
        if lr <= 0.0:
            raise ValueError('Invalid learning rate: {}'.format(lr))

        for i, beta in enumerate(betas):
            if not 0.0 <= beta < 1.0:
                msg = 'Invalid beta parameter at index 1: {}'.format(betas[i])
                raise ValueError(msg)

        if weight_decay < 0.0:
            raise ValueError(
                'Invalid weight_decay value: {}'.format(weight_decay)
            )

        defaults = dict(lr=lr, betas=betas, weight_decay=weight_decay)
        super(AggMo, self).__init__(params, defaults)

    @classmethod
    def from_exp_form(
        cls: typing.Type[T],
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        a: float=0.1,
        k: int=3,
        weight_decay: float=0,
    ) -> T:
        if lr <= 0.0:
            raise ValueError('Invalid parameter k: {}'.format(k))

        betas = [1 - a ** i for i in range(k)]  # type: List[float]
        return cls(params, lr, betas, weight_decay)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            betas = group['betas']
            total_mom = float(len(betas))

            for p in group['params']:
                if p.grad is None:
                    continue
                d_p = p.grad.data
                if weight_decay != 0:
                    d_p.add_(p.data, alpha=weight_decay)
                param_state = self.state[p]
                if 'momentum_buffer' not in param_state:
                    param_state['momentum_buffer'] = {}
                    for beta in betas:
                        param_state['momentum_buffer'][
                            beta
                        ] = torch.zeros_like(p.data)
                for beta in betas:
                    buf = param_state['momentum_buffer'][beta]
                    buf.mul_(beta).add_(d_p)
                    p.data.sub_(buf, alpha=group['lr'] / total_mom)
        return loss


from moai.validation.metrics.common.rmse import RMSE
from moai.validation.metrics.common.rmsle import RMSLE
from moai.validation.metrics.common.absrel import AbsRel
from moai.validation.metrics.common.sqrel import SqRel
from moai.validation.metrics.common.accuracy import (
    Accuracy,
    TopAccuracy,
    Top5Accuracy,
)

__all__ = [
    "RMSE",
    "RMSLE",
    "AbsRel",
    "SqRel",
    "Accuracy",
    "TopAccuracy",
    "Top5Accuracy",
]


try:
    from moai.action.train import train
    from moai.action.play import play
    from moai.action.evaluate import evaluate
    from moai.action.plot import plot
    from moai.action.diff import diff
    from moai.action.reprod import reprod
except:
    from action.train import train
    from action.play import play
    from action.evaluate import evaluate
    from action.plot import plot
    from action.diff import diff
    from action.reprod import reprod


import torch
import functools

__all__ = [
    "SphericalPad2d",
]

#NOTE: from https://github.com/pytorch/pytorch/issues/3858
#NOTE: check https://github.com/pytorch/pytorch/issues/20981
# also https://github.com/pytorch/pytorch/issues/3858#issuecomment-432801360

def __pad_circular_nd(x: torch.Tensor, pad: int, dim) -> torch.Tensor:
    """
    :param x: shape [H, W]
    :param pad: int >= 0
    :param dim: the dimension over which the tensors are padded
    :return:
    """
    if isinstance(dim, int):
        dim = [dim]
    for d in dim:
        if d >= len(x.shape):
            raise IndexError(f"dim {d} out of range")
        idx = tuple(slice(0, None if s != d else pad, 1) for s in range(len(x.shape)))
        x = torch.cat([x, x[idx]], dim=d)
        idx = tuple(slice(None if s != d else -2 * pad, None if s != d else -pad, 1) for s in range(len(x.shape)))
        x = torch.cat([x[idx], x], dim=d)
        pass
    return x

horizontal_circular_pad2d = functools.partial(__pad_circular_nd, dim=[3])

class SphericalPad2d(torch.nn.Module):
    def __init__(self,
        padding:        int = 1,
    ):
        super(SphericalPad2d, self).__init__()
        self.padding = padding
    
    def forward(self, x: torch.Tensor) -> torch.Tensor: # assumes [B, C, H, W] tensor inputs
        return torch.nn.functional.pad(
            horizontal_circular_pad2d(
                x, pad=self.padding
            ),
            pad=[0, 0, self.padding, self.padding], mode='replicate'
        )


from moai.data.datasets.common import load_color_image

import os
import torch
import glob
import typing
import logging
import json
import numpy as np
from collections import namedtuple

__all__ = [
    "OpenPoseInference",
]

logger = logging.getLogger(__name__)

OpenPoseParams = namedtuple('OpenPoseParams', [
    'load_hands',
    'load_face',
    'load_face_contour',
    'single_person_only',
])

class OpenPoseInferred(torch.utils.data.Dataset):
    def __init__(self,
        image_glob:                 str,
        load_hands:                 bool=False,
        load_face:                  bool=False,
        load_face_contour:          bool=False,
        single_person_only:         bool=False,
        invalid_joints:             typing.Sequence[int]=None,
        # updates:                    typing.Mapping[str, bool]={},
    ):
        super(OpenPoseInferred, self).__init__()
        self.params = OpenPoseParams(load_hands, 
            load_face, load_face_contour, single_person_only,
        )
        image_filenames = glob.glob(image_glob)
        self.filenames = image_filenames
        self.invalid_joints = list(invalid_joints or [])

    def __len__(self) -> int:
        return len(self.filenames)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        img_filename = self.filenames[index]
        folder, filename = os.path.split(img_filename)
        base_dir = os.path.basename(folder)
        base_dir = 'keypoints' if base_dir == 'images' else base_dir
        name, ext = os.path.splitext(filename)
        img = load_color_image(img_filename)
        keypoint_filename = os.path.join(os.path.dirname(folder), base_dir, f"{name}_keypoints.json")
        data = self._read_keypoints(keypoint_filename, 
            self.params.load_hands, self.params.load_face, self.params.load_face_contour
        )
        keypoints = data['keypoints']
        if self.params.single_person_only:
            def _get_area(keypoints: torch.Tensor) -> float:
                min_x = keypoints[..., 0].min()
                min_y = keypoints[..., 1].min()
                max_x = keypoints[..., 0].max()
                max_y = keypoints[..., 1].max()
                return (max_x - min_x) * (max_y - min_y) * keypoints[..., 2].sum()
            keypoints = [max(keypoints, key=_get_area)]
        keypoints = torch.stack(keypoints, dim=0).squeeze()
        keypoints[..., 2].scatter_(dim=0, 
            index=torch.Tensor(self.invalid_joints).long(), value=0.0
        )[:, np.newaxis]
        kpts = keypoints[..., :2]
        conf = keypoints[..., 2][:, np.newaxis]
        return {
            'color': img,
            'keypoints': kpts,
            'confidence': conf,
            'mask': (conf > 0.0).float()[:, np.newaxis]
        }

    def _read_keypoints(self,
        filename: str,
        load_hands=True,
        load_face=True,
        load_face_contour=False
    ) -> typing.Dict[str, torch.Tensor]:
        with open(filename) as keypoint_file:
            data = json.load(keypoint_file)
        keypoints, gender_pd, gender_gt = [], [], []
        for person in data['people']:
            body = np.array(person['pose_keypoints_2d'], dtype=np.float32)
            body = body.reshape([-1, 3])
            if load_hands:
                left_hand = np.array(person['hand_left_keypoints_2d'], dtype=np.float32).reshape([-1, 3])
                right_hand = np.array(person['hand_right_keypoints_2d'], dtype=np.float32).reshape([-1, 3])
                body = np.concatenate([body, left_hand, right_hand], axis=0)
            if load_face:
                face = np.array(person['face_keypoints_2d'], dtype=np.float32).reshape([-1, 3])[17: 17 + 51, :]
                contour_keyps = np.array([], dtype=body.dtype).reshape(0, 3)
                if load_face_contour:
                    contour_keyps = np.array(person['face_keypoints_2d'], dtype=np.float32).reshape([-1, 3])[:17, :]
                body = np.concatenate([body, face, contour_keyps], axis=0)

            gender_pd.append(person.get('gender_pd', None))
            gender_gt.append(person.get('gender_gt', None))
            keypoints.append(torch.from_numpy(body))
        return {
            'keypoints': keypoints,
            'gender_pd': gender_pd,
            'gender_gt': gender_gt, 
        }


from moai.parameters.optimization.schedule.schedulers import Identity

import torch
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["NoOp"]

class NoOp(object):
    def __init__(self,
        optimizers: typing.Sequence[torch.optim.Optimizer]
    ):
        self.schedulers = [Identity(next(iter(optimizers)))]

    def items(self):
        return self.schedulers


from collections import UserList
from pytorch_lightning import Callback

import  moai.utils.engine as mieng
import logging

log = logging.getLogger(__name__)

class EngineCallbacks(UserList):
    def __init__(self,
        **kwargs
    ):
        super().__init__()
        for k, v in kwargs.items():
            if isinstance(v, Callback):
                self.data.append(v)
            if isinstance(v, mieng.Collection):
                for i in v.items():
                    if isinstance(i, Callback):
                        self.data.append(i)


from collections.abc import Callable

import torch
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["Interval"]

class Interval(Callable):
    def __init__(self,
        batch_interval:int
    ):
        self.batch_interval = batch_interval

    @property
    def interval(self):
        return self.batch_interval


import moai.networks.lightning as minet
import moai.utils.parsing.rtp as mirtp

import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging
import inspect

log = logging.getLogger(__name__)

__all__ = ['Wrapper']

from moai.monads.execution.cascade import _create_accessor

class Wrapper(minet.FeedForward):
    def __init__(self,
        inner:          omegaconf.DictConfig,
        configuration:  omegaconf.DictConfig,
        data:           omegaconf.DictConfig=None,
        parameters:     omegaconf.DictConfig=None,
        feedforward:    omegaconf.DictConfig=None,
        monads:         omegaconf.DictConfig=None,
        supervision:    omegaconf.DictConfig=None,
        validation:     omegaconf.DictConfig=None,
        visualization:  omegaconf.DictConfig=None,
        export:         omegaconf.DictConfig=None,
    ):
        super(Wrapper, self).__init__(
            feedforward=feedforward, monads=monads, 
            visualization=visualization, data=data,
            supervision=supervision, validation=validation, 
            export=export, parameters=parameters,
        )         
        self.model = hyu.instantiate(inner)
        self.fwds = []
        params = inspect.signature(self.model.forward).parameters
        model_in = list(zip(*[mirtp.force_list(configuration.io[prop]) for prop in params]))
        model_out = mirtp.split_as(mirtp.resolve_io_config(configuration.io['out']), model_in)
        self.res_fill = [mirtp.get_result_fillers(self.model, out) for out in model_out]
        get_filler = iter(self.res_fill)
        for keys in model_in:
            accessors = [_create_accessor(k if isinstance(k, str) else k[0]) for k in keys]  
            self.fwds.append(lambda td,
                tk=keys,
                acc=accessors,
                args=params.keys(),
                model=self.model,
                filler=next(get_filler):
                    filler(td, model(**dict(zip(args,
                        list(acc[i](td) if type(k) is str
                                else list(td[j] for j in k)
                            for i, k in enumerate(tk)
                        )
                    ))))
            )
    
    def forward(self, 
        td: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        for f in self.fwds:
            f(td)
        return td
        


import torch
import typing

class Transformation(torch.nn.Module):
    __XYZ_AT__ = ['channel', 'row']

    def __init__(self,
        xyz_in_at:              str='channel', # ['channel', 'row]
        xyz_out_at:             str='channel', # ['channel', 'row]
        transpose:              bool=False,
    ):
        super(Transformation, self).__init__()
        self.transpose = transpose
        self.convert_points_in = self._convert_points_row_in if xyz_in_at == 'row'\
            else self._convert_points_channel_in
        self.convert_points_out = self._convert_points_row_out if xyz_out_at == 'row'\
            else self._convert_points_channel_out
        self.get_width_height = self._get_width_height_row_in if xyz_in_at == 'row'\
            else self._get_width_height_channel_in

    def _convert_points_row_in(self,
        points:                 torch.Tensor,
    ):
        return points.permute(0, 2, 1) if len(points.shape) == 3\
            else points.permute(0, 3, 1, 2)
        
    def _convert_points_channel_in(self,
        points:                 torch.Tensor,
    ):
        return points

    def _get_width_height_row_in(self,
        shape:                 torch.Size,
    ):
        return shape[2], shape[1]

    def _get_width_height_channel_in(self,
        shape:                 torch.Size,
    ):
        return shape[3], shape[2]

    def _convert_points_channel_out(self,
        points:                 torch.Tensor,
        shape:                  torch.Size,
    ):        
        return points if len(shape) == 3\
            else points.reshape(
                points.shape[0], *self.get_width_height(shape), 3
            )
        
    def _convert_points_row_out(self,
        points:                 torch.Tensor,
        shape:                  torch.Size,
    ):
        pts = points.transpose(2, 1)
        return pts if len(shape) == 3 else points.reshape(
            points.shape[0], 3, *self.get_width_height(shape)
        )

    def forward(self, 
        points:                 torch.Tensor,      # [B, 3, H, W] or [B, H, W, 3] or [B, 3, V] or [B, V, 3] 
        transform:              typing.Optional[torch.Tensor] = None, # [B, 4, 4]
        rotation:               typing.Optional[torch.Tensor] = None, # [B, 3, 3]
        translation:            typing.Optional[torch.Tensor] = None, # [B, 3]
    )-> torch.Tensor:
        b = points.shape[0]
        if transform is not None:
            xform = torch.transpose(transform, -1, -2) if self.transpose else transform
            rot = xform[:, :3, :3]
            #trans = transform[:, 3, :3].reshape(b, 3, 1) #TODO: What if the trans vector is in the last column
            trans = transform[:, :3, 3].reshape(b, 3, 1)
        else:
            rot = torch.transpose(rotation, 2, 1) if self.transpose else rotation
            trans = translation.reshape(b, 3, 1)
        pts3d = self.convert_points_in(points)
        xformed = rot @ pts3d.view(b, 3, -1) + trans
        return self.convert_points_out(xformed, points.size())
        



from moai.supervision.losses.regression import L2

import torch

#NOTE: adapted from https://github.com/vchoutas/smplify-x
#NOTE: Statistical methods for tomographic image reconstruction, S Geman,  D McClure

__all__ = ["GemanMcClure"]

class GemanMcClure(L2):
    r"""Implements the Geman-McClure error function.

    """
    def __init__(self,
        rho: float=1.0,
    ):
        super(GemanMcClure, self).__init__()
        self.rho_sq = rho ** 2

    def forward(self,
        pred:       torch.Tensor,
        gt:         torch.Tensor,        
        weights:    torch.Tensor=None, # float tensor
        mask:       torch.Tensor=None, # byte tensor
    ) -> torch.Tensor:
        L2 = super(GemanMcClure, self).forward(pred=pred, gt=gt)
        gm = L2 / (L2 + self.rho_sq) * self.rho_sq
        if weights is not None:
            gm = gm * weights
        if mask is not None:
            gm = gm[mask]
        return gm


import sys
import hydra
import os
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ['Import']

class Import(object):
    def __init__(self,
        run_path:               bool=True,
        other_paths:            typing.Sequence[str]=[],
    ):
        if run_path:
            cwd = hydra.utils.get_original_cwd()
            sys.path.append(cwd)
            log.info(f"Adding current working directory [{os.path.abspath(cwd)}] to path.")
        for path in other_paths:
            sys.path.append(path)
            log.info(f"Adding [{os.path.abspath(path)}] to path.")


import typing
import torch
import math
import logging

log = logging.getLogger(__name__)

__all__ = [
    "CentralizedRanger",
    "CentralizedRAdam",
    "CentralizedPlainRAdam",
    "CentralizedSGD",
    "CentralizedAdam",
    "CentralizedAdamW",
]

#NOTE: from https://github.com/Yonghongwei/Gradient-Centralization/tree/ce61b65dd4cd60074738f6acf00cf3ec98d10079/algorithm-GC/algorithm

def centralized_gradient(x,use_gc=True,gc_conv_only=False):
    if use_gc:
      if gc_conv_only:
        if len(list(x.size()))>3:
            x.add_(-x.mean(dim = tuple(range(1,len(list(x.size())))), keepdim = True))
      else:
        if len(list(x.size()))>1:
            x.add_(-x.mean(dim = tuple(range(1,len(list(x.size())))), keepdim = True))
    return x 

class CentralizedSGD(torch.optim.Optimizer):
    r"""Implements the Centralized Gradient algorithm with SGD.

    - **Paper**: [Gradient Centralization: A New Optimization Technique for Deep Neural Networks]https://arxiv.org/pdf/2004.01461.pdf)
    - **Implementation**: [GitHub @ Yonghongwei](https://github.com/Yonghongwei/Gradient-Centralization)

    Nesterov momentum is based on the formula from
    `On the importance of initialization and momentum in deep learning`__.

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)

    Example:
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.1, momentum=0.9)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ http://www.cs.toronto.edu/%7Ehinton/absps/momentum.pdf

    .. note::
        The implementation of SGD with Momentum/Nesterov subtly differs from
        Sutskever et. al. and implementations in some other frameworks.

        Considering the specific case of Momentum, the update can be written as

        .. math::
            \begin{aligned}
                v_{t+1} & = \mu * v_{t} + g_{t+1}, \\
                p_{t+1} & = p_{t} - \text{lr} * v_{t+1},
            \end{aligned}

        where :math:`p`, :math:`g`, :math:`v` and :math:`\mu` denote the 
        parameters, gradient, velocity, and momentum respectively.

        This is in contrast to Sutskever et. al. and
        other frameworks which employ an update of the form

        .. math::
            \begin{aligned}
                v_{t+1} & = \mu * v_{t} + \text{lr} * g_{t+1}, \\
                p_{t+1} & = p_{t} - v_{t+1}.
            \end{aligned}

        The Nesterov version is analogously modified.
    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        #lr: float=torch.optim.optimizer.required,
        lr=None,
        momentum: float=0,
        dampening: float=0,
        weight_decay: float=0,
        nesterov: bool=False,
        use_gc: bool=True,
        gc_conv_only: bool=False
    ):
        #if lr is not torch.optim.optimizer.required and lr < 0.0:
        if lr is not None and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if momentum < 0.0:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = dict(lr=lr, momentum=momentum, dampening=dampening,
                        weight_decay=weight_decay, nesterov=nesterov, use_gc=use_gc,gc_conv_only=gc_conv_only)
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")
        super(CentralizedSGD, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(CentralizedSGD, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)

    @torch.no_grad()
    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']

            for p in group['params']:
                if p.grad is None:
                    continue
                d_p = p.grad
                if weight_decay != 0:
                    d_p = d_p.add(p, alpha=weight_decay)
                    
                #GC operation     
                d_p =centralized_gradient(d_p ,use_gc=group['use_gc'],gc_conv_only=group['gc_conv_only'])                 
                                        
                if momentum != 0:
                    param_state = self.state[p]
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.clone(d_p).detach()
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(d_p, alpha=1 - dampening)
                    if nesterov:
                        d_p = d_p.add(buf, alpha=momentum)
                    else:
                        d_p = buf
 
                             
                p.add_(d_p, alpha=-group['lr'])

        return loss

class CentralizedAdam(torch.optim.Optimizer):
    r"""Implements the Centralized Gradient algorithm with Adam.

    - **Paper**: [Gradient Centralization: A New Optimization Technique for Deep Neural Networks]https://arxiv.org/pdf/2004.01461.pdf)
    - **Implementation**: [GitHub @ Yonghongwei](https://github.com/Yonghongwei/Gradient-Centralization)

    It has been proposed in `Adam: A Method for Stochastic Optimization`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper `On the Convergence of Adam and Beyond`_
            (default: False)

    .. _Adam\: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ
    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-8,
        weight_decay: float=0,
        amsgrad: bool=False,
        use_gc: bool=True,
        gc_conv_only: bool=False,
        gc_loc: bool=False
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay, amsgrad=amsgrad)
        super(CentralizedAdam, self).__init__(params, defaults)
        self.gc_loc=gc_loc
        self.use_gc=use_gc
        self.gc_conv_only=gc_conv_only

    def __setstate__(self, state):
        super(CentralizedAdam, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsgrad', False)

    @torch.no_grad()
    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')
                amsgrad = group['amsgrad']

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    if amsgrad:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = torch.zeros_like(p, memory_format=torch.preserve_format)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if amsgrad:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                if group['weight_decay'] != 0:
                    grad = grad.add(p, alpha=group['weight_decay'])
                if self.gc_loc:
                   grad=centralized_gradient(grad,use_gc=self.use_gc,gc_conv_only=self.gc_conv_only)
                    
                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                if amsgrad:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = (max_exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])
                else:
                    denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])

                step_size = group['lr'] / bias_correction1
                #GC operation 
                G_grad=exp_avg/denom 
                if self.gc_loc==False:       
                    G_grad=centralized_gradient(G_grad,use_gc=self.use_gc,gc_conv_only=self.gc_conv_only)
                
                p.add_( G_grad, alpha=-step_size)

        return loss

class CentralizedAdamW(torch.optim.Optimizer):
    r"""Implements the Centralized Gradient algorithm with AdamW.

    - **Paper**: [Gradient Centralization: A New Optimization Technique for Deep Neural Networks]https://arxiv.org/pdf/2004.01461.pdf)
    - **Implementation**: [GitHub @ Yonghongwei](https://github.com/Yonghongwei/Gradient-Centralization)

    The original Adam algorithm was proposed in `Adam: A Method for Stochastic Optimization`_.
    The AdamW variant was proposed in `Decoupled Weight Decay Regularization`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay coefficient (default: 1e-2)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper `On the Convergence of Adam and Beyond`_
            (default: False)

    .. _Adam\: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _Decoupled Weight Decay Regularization:
        https://arxiv.org/abs/1711.05101
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ
    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-8,
        weight_decay: float=1e-2,
        amsgrad: bool=False,
        use_gc: bool=True,
        gc_conv_only: bool=False,
        gc_loc: bool=True
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay, amsgrad=amsgrad)
        super(CentralizedAdamW, self).__init__(params, defaults)
        self.gc_loc=gc_loc
        self.use_gc=use_gc
        self.gc_conv_only=gc_conv_only
        
    def __setstate__(self, state):
        super(CentralizedAdamW, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsgrad', False)

    @torch.no_grad()
    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                # Perform optimization step
                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError('AdamW does not support sparse gradients')
                amsgrad = group['amsgrad']

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    if amsgrad:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = torch.zeros_like(p, memory_format=torch.preserve_format)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if amsgrad:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                if self.gc_loc:
                   grad=centralized_gradient(grad,use_gc=self.use_gc,gc_conv_only=self.gc_conv_only)

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                if amsgrad:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = (max_exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])
                else:
                    denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(group['eps'])

                step_size = group['lr'] / bias_correction1

                #GC operation and stepweight decay
                G_grad=(exp_avg/denom).add(p.data,alpha=group['weight_decay'])        
                if self.gc_loc==False:       
                    G_grad=centralized_gradient(G_grad,use_gc=self.use_gc,gc_conv_only=self.gc_conv_only)
                
                p.add_( G_grad, alpha=-step_size)

        return loss

class CentralizedRAdam(torch.optim.Optimizer):
    """Implements the Centralized Gradient algorithm for RAdam.

    - **Paper**: [Gradient Centralization: A New Optimization Technique for Deep Neural Networks]https://arxiv.org/pdf/2004.01461.pdf)
    - **Implementation**: [GitHub @ Yonghongwei](https://github.com/Yonghongwei/Gradient-Centralization)

    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        alpha: float=0.5,
        k: int=6,
        N_sma_threshhold: int=5,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-8,
        weight_decay: float=0,  # Adam options
        degenerated_to_sgd: bool=True,
        # Gradient centralization on or off, applied to conv layers only or conv + fc layers
        use_gc=True,
        gc_conv_only=False,
        gc_loc=False
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))

        self.degenerated_to_sgd = degenerated_to_sgd
        self.gc_loc=gc_loc
        self.use_gc=use_gc
        self.gc_conv_only=gc_conv_only
                
        if isinstance(params, (list, tuple)) and len(params) > 0 and isinstance(params[0], dict):
            for param in params:
                if 'betas' in param and (param['betas'][0] != betas[0] or param['betas'][1] != betas[1]):
                    param['buffer'] = [[None, None, None] for _ in range(10)]
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, buffer=[[None, None, None] for _ in range(10)])
        super(CentralizedRAdam, self).__init__(params, defaults)

        
    def __setstate__(self, state):
        super(CentralizedRAdam, self).__setstate__(state)

    def step(self, closure=None):

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError('RAdam does not support sparse gradients')

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']
                if self.gc_loc:
                   grad=centralized_gradient(grad,use_gc=self.use_gc,gc_conv_only=self.gc_conv_only)
                   
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                


                state['step'] += 1
                buffered = group['buffer'][int(state['step'] % 10)]
                if state['step'] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state['step']
                    beta2_t = beta2 ** state['step']
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * state['step'] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma

                    # more conservative since it's an approximated value
                    if N_sma >= 5:
                        step_size = math.sqrt((1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (N_sma - 2) / N_sma * N_sma_max / (N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    elif self.degenerated_to_sgd:
                        step_size = 1.0 / (1 - beta1 ** state['step'])
                    else:
                        step_size = -1
                    buffered[2] = step_size

                # more conservative since it's an approximated value
                if N_sma >= 5:
                    denom = exp_avg_sq.sqrt().add_(group['eps'])                                        
                    G_grad=exp_avg/denom  
                elif step_size > 0:
                    G_grad=exp_avg  
       
                if group['weight_decay'] != 0:
                       G_grad.add_(p_data_fp32,alpha=group['weight_decay'])   
                #GC operation                                                  
                if self.gc_loc==False:       
                    G_grad=centralized_gradient(G_grad,use_gc=self.use_gc,gc_conv_only=self.gc_conv_only)                                       
                p_data_fp32.add_( G_grad, alpha=-step_size * group['lr'])
                    #p_data_fp32.addcdiv_(-step_size * group['lr'], exp_avg, denom)
                p.data.copy_(p_data_fp32)
        return loss

class CentralizedPlainRAdam(torch.optim.Optimizer):
    """Implements the Centralized Gradient algorithm for RAdam.

    - **Paper**: [Gradient Centralization: A New Optimization Technique for Deep Neural Networks]https://arxiv.org/pdf/2004.01461.pdf)
    - **Implementation**: [GitHub @ Yonghongwei](https://github.com/Yonghongwei/Gradient-Centralization)

    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        alpha: float=0.5,
        k: int=6,
        N_sma_threshhold: int=5,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-8,
        weight_decay: float=0,  # Adam options
        degenerated_to_sgd: bool=True,
        # Gradient centralization on or off, applied to conv layers only or conv + fc layers
        use_gc=True,
        gc_conv_only=False,
        gc_loc=False
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
                    
        self.degenerated_to_sgd = degenerated_to_sgd
        self.gc_loc=gc_loc
        self.use_gc = use_gc
        self.gc_conv_only=gc_conv_only
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)

        super(CentralizedPlainRAdam, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(CentralizedPlainRAdam, self).__setstate__(state)

    def step(self, closure=None):

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError('RAdam does not support sparse gradients')

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']
                if self.gc_loc:
                   grad=centralized_gradient(grad,use_gc=self.use_gc,gc_conv_only=self.gc_conv_only)
                   
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

                state['step'] += 1
                beta2_t = beta2 ** state['step']
                N_sma_max = 2 / (1 - beta2) - 1
                N_sma = N_sma_max - 2 * state['step'] * beta2_t / (1 - beta2_t)


                # more conservative since it's an approximated value
                if N_sma >= 5:
                    #if group['weight_decay'] != 0:
                    #    p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)
                    step_size = group['lr'] * math.sqrt((1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (N_sma - 2) / N_sma * N_sma_max / (N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    denom = exp_avg_sq.sqrt().add_(group['eps'])                                       
                    G_grad=exp_avg/denom  

                elif self.degenerated_to_sgd:
                    #if group['weight_decay'] != 0:
                    #    p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)
                    step_size = group['lr'] / (1 - beta1 ** state['step'])
                    G_grad=exp_avg  

                if group['weight_decay'] != 0:
                       G_grad.add_(p.data,alpha=group['weight_decay']) 
                                                                   
                if self.gc_loc==False:       
                    G_grad=centralized_gradient(G_grad,use_gc=self.use_gc,gc_conv_only=self.gc_conv_only)
                    
                p_data_fp32.add_( G_grad, alpha=-step_size * group['lr'])                   
                    #p_data_fp32.addcdiv_(-step_size, exp_avg, denom)
                p.data.copy_(p_data_fp32)
        return loss

class CentralizedRanger(torch.optim.Optimizer):
    """Implements the Centralized Gradient algorithm for Ranger.

    - **Paper**: [Gradient Centralization: A New Optimization Technique for Deep Neural Networks]https://arxiv.org/pdf/2004.01461.pdf)
    - **Implementation**: [GitHub @ Yonghongwei](https://github.com/Yonghongwei/Gradient-Centralization)

    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        alpha: float=0.5,
        k: int=6,
        N_sma_threshhold: int=5,
        betas: typing.Tuple[float, float]=(.95, 0.999),
        eps: float=1e-5,
        weight_decay: float=0,  # Adam options
        # Gradient centralization on or off, applied to conv layers only or conv + fc layers
        use_gc=True,
        gc_conv_only=False,
        gc_loc=False
    ):
        # parameter checks
        if not 0.0 <= alpha <= 1.0:
            raise ValueError(f'Invalid slow update rate: {alpha}')
        if not 1 <= k:
            raise ValueError(f'Invalid lookahead steps: {k}')
        if not lr > 0:
            raise ValueError(f'Invalid Learning Rate: {lr}')
        if not eps > 0:
            raise ValueError(f'Invalid eps: {eps}')

        # parameter comments:
        # beta1 (momentum) of .95 seems to work better than .90...
        # N_sma_threshold of 5 seems better in testing than 4.
        # In both cases, worth testing on your dataset (.90 vs .95, 4 vs 5) to make sure which works best for you.

        # prep defaults and init torch.optim base
        defaults = dict(lr=lr, alpha=alpha, k=k, step_counter=0, betas=betas,
                        N_sma_threshhold=N_sma_threshhold, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)

        # adjustable threshold
        self.N_sma_threshhold = N_sma_threshhold

        # look ahead params

        self.alpha = alpha
        self.k = k

        # radam buffer for state
        self.radam_buffer = [[None, None, None] for ind in range(10)]

        # gc on or off
        self.gc_loc=gc_loc
        self.use_gc = use_gc
        self.gc_conv_only=gc_conv_only
        # level of gradient centralization
        #self.gc_gradient_threshold = 3 if gc_conv_only else 1

        print(f"Ranger optimizer loaded. \nGradient Centralization usage = {self.use_gc}")
        if (self.use_gc and self.gc_conv_only == False):
            print(f"GC applied to both conv and fc layers")
        elif (self.use_gc and self.gc_conv_only == True):
            print(f"GC applied to conv layers only")

    def __setstate__(self, state):
        super(CentralizedRanger, self).__setstate__(state)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        loss = None
        # note - below is commented out b/c I have other work that passes back the loss as a float, and thus not a callable closure.
        # Uncomment if you need to use the actual closure...

        # if closure is not None:
        #loss = closure()

        # Evaluate averages and grad, update param tensors
        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()

                if grad.is_sparse:
                    raise RuntimeError(
                        'Ranger optimizer does not support sparse gradients')

                p_data_fp32 = p.data.float()

                state = self.state[p]  # get state dict for this param

                if len(state) == 0:  # if first time to run...init dictionary with our desired entries
                    # if self.first_run_check==0:
                    # self.first_run_check=1
                    #print("Initializing slow buffer...should not see this at load from saved model!")
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)

                    # look ahead weight storage now in state dict
                    state['slow_buffer'] = torch.empty_like(p.data)
                    state['slow_buffer'].copy_(p.data)

                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(
                        p_data_fp32)

                # begin computations
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                # GC operation for Conv layers and FC layers
                #if grad.dim() > self.gc_gradient_threshold:
                #    grad.add_(-grad.mean(dim=tuple(range(1, grad.dim())), keepdim=True))
                if self.gc_loc:
                   grad=centralized_gradient(grad,use_gc=self.use_gc,gc_conv_only=self.gc_conv_only)
                   
                state['step'] += 1

                # compute variance mov avg
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                
                # compute mean moving avg
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                

                buffered = self.radam_buffer[int(state['step'] % 10)]

                if state['step'] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state['step']
                    beta2_t = beta2 ** state['step']
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * \
                        state['step'] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma
                    if N_sma > self.N_sma_threshhold:
                        step_size = math.sqrt((1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (
                            N_sma - 2) / N_sma * N_sma_max / (N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    else:
                        step_size = 1.0 / (1 - beta1 ** state['step'])
                    buffered[2] = step_size

                #if group['weight_decay'] != 0:
                #    p_data_fp32.add_(-group['weight_decay']
                #                     * group['lr'], p_data_fp32)

                # apply lr
                if N_sma > self.N_sma_threshhold:
                    denom = exp_avg_sq.sqrt().add_(group['eps'])                                         
                    G_grad=exp_avg/denom                                                                                    
                else:
                    G_grad=exp_avg  

                if group['weight_decay'] != 0:
                       G_grad.add_(p_data_fp32,alpha=group['weight_decay']) 
                #GC operation                                                   
                if self.gc_loc==False:       
                    G_grad=centralized_gradient(G_grad,use_gc=self.use_gc,gc_conv_only=self.gc_conv_only)
                                                           
                p_data_fp32.add_( G_grad, alpha=-step_size * group['lr'])
                p.data.copy_(p_data_fp32)

                # integrated look ahead...
                # we do it at the param level instead of group level
                if state['step'] % group['k'] == 0:
                    # get access to slow param tensor
                    slow_p = state['slow_buffer']
                    # (fast weights - slow weights) * alpha
                    slow_p.add_( p.data - slow_p,alpha=self.alpha)
                    # copy interpolated weights to RAdam param tensor
                    p.data.copy_(slow_p)

        return loss


__version__ = '0.1.5a1'
__author__ = 'AI-in-Motion Team'
__author_email__ = 'moai@ai-in-motion.dev'
__license__ = 'Apache-2.0'
__copyright__ = f'Copyright (c) 2022, {__author__}.' 
__homepage__ = 'https://github.com/ai-in-motion/moai'
__github_repo__ = 'https://github.com/ai-in-motion/moai'
__documentation_page__ = 'https://moai.readthedocs.io/en/latest/'
__docs__ = "moai: Accelerated, Flexible, Modular, Reproducible, Insightful AI"
__keywords__ = 'AI,workflow,framework,PyTorch'


import torch
import functools
import logging

log = logging.getLogger(__name__)

__all__ = [
    "Adaptive",
]

class Adaptive(torch.nn.Module):
    def __init__(self,
        scale_factor:       float=2.0,  # downscale factor, divides the resolution
        mode:               str='max',  # 'max' or 'avg' pool
        dims:               int=2,      # 1, 2 or 3 dimensional downsampling
    ):
        super(Adaptive, self).__init__()
        self.pool_func = getattr(torch.nn.functional, f"adaptive_{mode}_pool{dims}d")
        self.scale_factor = scale_factor
        self.dims = dims

    def even_size(self, size: int, scale_factor: float) -> int:
        downscaled = int(size // scale_factor)
        return downscaled + int(downscaled % 2)

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        size = [self.even_size(s, self.scale_factor) for s in tensor.shape[2:2+ self.dims]]
        return self.pool_func(tensor, size)

Adaptive1d = functools.partial(Adaptive, dims=1)
Adaptive2d = functools.partial(Adaptive, dims=2)
Adaptive3d = functools.partial(Adaptive, dims=3)

AdaptiveMaxPool1d = functools.partial(Adaptive, dims=1, mode='max')
AdaptiveMaxPool2d = functools.partial(Adaptive, dims=2, mode='max')
AdaptiveMaxPool3d = functools.partial(Adaptive, dims=3, mode='max')

AdaptiveAvgPool1d = functools.partial(Adaptive, dims=1, mode='avg')
AdaptiveAvgPool2d = functools.partial(Adaptive, dims=2, mode='avg')
AdaptiveAvgPool3d = functools.partial(Adaptive, dims=3, mode='avg')



import torch
import toolz
import typing

__all__ = ['get_submodule']

def get_submodule_pt_ge_110(
    module: torch.nn.Module,
    name: str
) -> torch.nn.Module:
    return module.get_submodule(name)

def get_submodule_pt_lt_110(
    module: torch.nn.Module,
    name: str
) -> torch.nn.Module:
    split = name.split('.')
    def _getattr(object: typing.Any, key: str):
        return getattr(object, key, None)
    return toolz.reduce(_getattr, split, module)

if isinstance(torch.__version__, str):
    v = torch.__version__.split('.')
else:
    v = torch.__version__
if (int(v[0]), int(v[1])) >= (1, 10):
    get_submodule = get_submodule_pt_ge_110
else:
    get_submodule = get_submodule_pt_lt_110


from moai.monads.masking.binary import Binary
from moai.monads.masking.finite import Finite
from moai.monads.masking.mask import Mask, Index
from moai.monads.masking.threshold import Threshold

__all__ = [    
    "Binary",
    "Finite",
    "Mask",
    "Threshold",
    "Index",
]


from moai.utils.arguments import (
    assert_numeric,
    assert_choices
)

from kornia.filters import (
    Laplacian,
    SpatialGradient,
)

import torch
import logging
import functools

log = logging.getLogger(__name__)

__all__ = ["Laplacian2d", "Sobel2d", "Diff2d"]

class Image2d(SpatialGradient):

    __MODES__ = ['sobel', 'diff']
    __REDUCTIONS__ = ['none', 'magnitude', 'absum']

    def __init__(self,
        mode:               str='sobel', # one of ['sobel', 'diff']
        order:              int=1,
        normalized:         bool=True,
        reduction:          str='magnitude',        
    ):
        assert_numeric(log, "order", order, min_value=1, max_value= 2)
        assert_choices(log, "mode", mode, Image2d.__MODES__)
        assert_choices(log, "reduction", reduction, Image2d.__REDUCTIONS__)
        super(Image2d, self).__init__(mode=mode, order=order, normalized=normalized)        
        def abs_sum(t: torch.Tensor, dim: int) -> torch.Tensor:
            return torch.sum(t.abs(), dim=[2])
        self.reduction_func = functools.partial(abs_sum, dim=2)\
            if reduction == 'abssum' else (
                functools.partial(torch.linalg.norm, ord=2, dim=2)
                if reduction == 'magnitude' else lambda t: t
            )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        filtered = super(Image2d, self).forward(image)
        return self.reduction_func(filtered)

Sobel2d = functools.partial(Image2d, mode='sobel')
Diff2d = functools.partial(Image2d, mode='diff')

class Laplacian2d(Laplacian):
    def __init__(self,
        kernel_size:                int=5,
        border_type:                str='reflect',
        normalized:                 bool=True,
    ):
        super(Laplacian2d, self).__init__(kernel_size=kernel_size,
            border_type=border_type, normalized=normalized
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return super(Laplacian2d, self).forward(image)



from moai.visualization.visdom.base import Base
from moai.utils.color.colorize import get_colormap, COLORMAPS
from moai.utils.arguments import assert_numeric

import torch
import visdom
import functools
import typing
import logging
import numpy as np
import math

log = logging.getLogger(__name__)

__all__ = ["Image2d"]

class Image2d(Base):
    def __init__(self,
        image:             typing.Union[str, typing.Sequence[str]],
        type:              typing.Union[str, typing.Sequence[str]],
        colormap:          typing.Union[str, typing.Sequence[str]],
        transform:         typing.Union[str, typing.Sequence[str]],
        batch_percentage:   float=1.0,
        name:               str="default",
        ip:                 str="http://localhost",
        port:               int=8097,   
    ):
        super(Image2d, self).__init__(name, ip, port)
        self.keys = [image] if isinstance(image, str) else list(image)
        self.types = [type] if isinstance(type, str) else list(type)
        self.transforms = [transform] if isinstance(transform, str) else list(transform)
        self.colormaps = [colormap] if isinstance(colormap, str) else list(colormap)
        self.batch_percentage = batch_percentage
        assert_numeric(log, 'batch percentage', batch_percentage, 0.0, 1.0)
        self.viz_map = {
            'color': functools.partial(self._viz_color, self.visualizer),
            'heatmap': functools.partial(self._viz_heatmap, self.visualizer),
        }
        self.transform_map = {
            'none': functools.partial(self._passthrough),            
            'minmax': functools.partial(self._minmax_normalization),
            'ndc': functools.partial(self._ndc_to_one),
            'dataset': functools.partial(self._dataset_normalization),
        }
        self.colorize_map = { "none": lambda x: x }
        self.colorize_map.update(COLORMAPS)

    @property
    def name(self) -> str:
        return self.env_name
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for k, t, tf, c in zip(self.keys, self.types, self.transforms, self.colormaps):
            self.viz_map[t](
                self.colorize_map[c](
                    self.transform_map[tf](
                        tensors, k, int(math.ceil(self.batch_percentage * tensors[k].shape[0])),
                        # tensors[k][:int(self.batch_percentage * tensors[k].shape[0])]
                    )
                ), k, k, self.name
            )

    @staticmethod
    def _viz_color(
        visdom: visdom.Visdom,
        array: np.array,
        key: str,
        win: str,
        env: str
    ) -> None:
        visdom.images(
            array,
            win=win,
            env=env,
            opts={
                'title': key,
                'caption': key,
                'jpgquality': 50,
            }
        )

    @staticmethod
    def _viz_heatmap(
        visdom: visdom.Visdom,
        tensor: torch.Tensor,
        key: str,
        win: str,
        env: str
    ) -> None:
        b, _, __, ___ = tensor.size() #NOTE: assumes [B, C, H, W], i.e. 2d train
        heatmaps = torch.flip(tensor, dims=[2]).detach().cpu()
        for i in range(b):
            opts = (
            {
                'title': key + "_" + str(i),
                'colormap': 'Viridis'
            })
            visdom.heatmap(heatmaps[i, :, :, :].squeeze(), opts=opts, win=win + str(i))

    @staticmethod #TODO: refactor these into a common module
    def _passthrough(
        tensors:    typing.Dict[str, torch.Tensor],
        key:        str,
        count:      int,
    ) -> torch.Tensor:
        return tensors[key][:count] if key in tensors else None

    @staticmethod #TODO: refactor these into a common module
    def _minmax_normalization(
        tensors:    typing.Dict[str, torch.Tensor],
        key:        str,
        count:      int,
    ) -> torch.Tensor:
        tensor = Image2d._passthrough(tensors, key, count)
        b, _, __, ___ = tensor.size()
        min_v = torch.min(tensor.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        max_v = torch.max(tensor.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        return (tensor - min_v) / (max_v - min_v)

    @staticmethod #TODO: refactor these into a common module
    def _ndc_to_one(
        tensors:    typing.Dict[str, torch.Tensor],
        key:        str,
        count:      int,
    ) -> torch.Tensor:
        tensor = Image2d._passthrough(tensors, key, count)
        return torch.addcmul(
            torch.scalar_tensor(0.5).to(tensor), 
            tensor,
            torch.scalar_tensor(0.5).to(tensor)
        )

    @staticmethod #TODO: refactor these into a common module
    def _dataset_normalization(
        tensors:    typing.Dict[str, torch.Tensor],
        key:        str,
        count:      int,
    ) -> torch.Tensor:
        tensor = Image2d._passthrough(tensors, key, count)
        mean = Image2d._passthrough(tensors, 'dataset_mean', count)
        if mean is None:
            mean = torch.scalar_tensor(0.0).to(tensor)
        std = Image2d._passthrough(tensors, 'dataset_std', count)
        if std is None:
            std = torch.scalar_tensor(1.0).to(tensor)
        return torch.addcmul(
            mean, 
            tensor,
            std
        )


from moai.parameters.optimization.single import Optimizer as Single
from moai.parameters.optimization.noop import NoOp
from moai.parameters.optimization.lookahead import Lookahead
from moai.parameters.optimization.larc import LARC
from moai.parameters.optimization.swa import SWA

__all__ = [
    "NoOp",
    "Single",
    "Lookahead",
    "LARC",
    "SWA",
]


import moai.networks.lightning as minet
import moai.utils.parsing.rtp as mirtp

from collections import OrderedDict

import torch
import inspect
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging

__all__ = ['Cascade']

log = logging.getLogger(__name__)

def _create_processing_block(
    cfg: omegaconf.DictConfig, 
    attribute: str, 
    monads: omegaconf.DictConfig
):
    if not cfg and attribute in cfg:
        log.warning(f"Empty processing block ({attribute}) in feedforward model.")
    return hyu.instantiate(getattr(cfg, attribute), monads)\
        if cfg and attribute in cfg else torch.nn.Identity()

class Cascade(minet.FeedForward):
    def __init__(self,
        configuration:  omegaconf.DictConfig, 
        modules:        omegaconf.DictConfig,        
        data:           omegaconf.DictConfig=None,
        parameters:     omegaconf.DictConfig=None,
        feedforward:    omegaconf.DictConfig=None,
        monads:         omegaconf.DictConfig=None,
        supervision:    omegaconf.DictConfig=None,
        validation:     omegaconf.DictConfig=None,
        visualization:  omegaconf.DictConfig=None,
        export:         omegaconf.DictConfig=None,
    ):
        super(Cascade, self).__init__(
            data=data, parameters=parameters,
            feedforward=feedforward, monads=monads,
            supervision=supervision, validation=validation,
            visualization=visualization, export=export,
        )

        self.mods = torch.nn.ModuleDict(OrderedDict(
            (k, hyu.instantiate(d)) for k, d in sorted(
                modules.items(), 
                key=lambda kvp: kvp[0]
            )
        ))

        self.mod_fwds = OrderedDict()
        for name, module in self.mods.items():
            params = inspect.signature(module.forward).parameters
            mod_in = list(zip(*[mirtp.force_list(configuration[name][prop]) for prop in params if configuration[name][prop] is not None]))
            #NOTE: mod_in receives them ordered (list), and so args cannot be ignored (nulled) when not last in the list
            #TODO: fix it via mapping
            mod_out = mirtp.split_as(mirtp.resolve_io_config(configuration[name]['out']), mod_in)

            mod_res_fill = [mirtp.get_result_fillers(module, out) for out in mod_out]
            get_mod_filler = iter(mod_res_fill)
            
            self.mod_fwds[name] = []
            for keys in mod_in: #TODO: check if feedforward to ignore filler
                self.mod_fwds[name].append(lambda td,
                    tk=keys,
                    args=params.keys(), 
                    mod=module,
                    filler=next(get_mod_filler):
                        filler(td, mod(**dict(zip(args,
                            list(
                                    td[k] if type(k) is str
                                    else list(td[j] for j in k)
                                for k in tk
                            )
                        ))))
                )
        self.pre_module = torch.nn.ModuleList(
            _create_processing_block(feedforward, f'pre_module{i}', monads)\
                for i in range(1, len(self.mods) + 1)
        )

    def forward(self, 
        td: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        for ((k, v), pre) in zip(
            self.mod_fwds.items(),
            self.pre_module,
        ):
            for d in v:
                td = pre(td)
                d(td)
        return td


from moai.utils.arguments import (
    ensure_path,
    ensure_choices,
)

import pickle
import torch
import typing
import logging
import os
import toolz

__all__ = ["Pkl"]

log = logging.getLogger(__name__)

class Pkl(typing.Callable[[typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]], None]):

    __MODES__ = ['all', 'append']
    
    def __init__(self,
        path:           str,
        keys:           typing.Union[str, typing.Sequence[str]],
        mode:           str="append", # all"
        counter_format: str="05d",
    ):
        self.mode = ensure_choices(log, "saving mode", mode, Pkl.__MODES__)
        self.folder = ensure_path(log, "output folder", path)
        self.index = 0
        self.keys = [keys] if type(keys) is str else list(keys)
        self.fmt = counter_format
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        arrays = { }
        for key in self.keys:
            split = key.split('.')
            arrays[key] = toolz.get_in(split, tensors).detach().cpu().numpy()
        if self.mode == 'append':
            mode = 'ab'
            batch = toolz.get_in(['__moai__', 'batch_index'], tensors)
            step = toolz.get_in(['__moai__', 'optimization_step'], tensors)
            stage = toolz.get_in(['__moai__', 'optimization_stage'], tensors)
            save = { 
                'optimization_state': {
                    'iteration': str(step),
                    'stage': stage,
                }, 'parameters_state': arrays
            } if step else arrays
            with open(os.path.join(self.folder, f"results_{batch:{self.fmt}}.pkl"), mode) as f:
                pickle.dump(save, f)
        else:
            mode = 'b'
            log.error("Pickle exporting is not yet enabled in non append mode.")
        


from collections.abc import Callable
from moai.utils.arguments import (
    ensure_path,
    ensure_choices,
)

import moai.utils.color.colorize as mic

import os
import torch
import torchvision
import visdom
import numpy
import functools
import typing
import logging
import cv2

log = logging.getLogger(__name__)

__all__ = ["Blend2d"]

class Blend2d(Callable):

    __MODES__ = ['overwrite', 'all']
    __FORMATS__ = ['jpg', 'png', 'exr']

    def __init__(self,
        path:           str,
        left:           typing.Union[str, typing.Sequence[str]],
        right:          typing.Union[str, typing.Sequence[str]],
        blending:       typing.Union[float, typing.Sequence[float]],
        colormap:       typing.Union[str, typing.Sequence[str]],
        transform:      typing.Union[str, typing.Sequence[str]],
        mode:           str="overwrite", # all"
        extensions:     typing.Union[str, typing.Sequence[str]]=["png"], # jpg or png or exr
        scale:          float=1.0,
    ):
        self.mode = ensure_choices(log, "saving mode", mode, Blend2d.__MODES__)
        folder = ensure_path(log, "output folder", path)
        self.folder = os.path.join(os.getcwd(), folder)
        self.formats = [ensure_choices(log, "output format", ext, Blend2d.__FORMATS__) for ext in extensions]
        self.mode = mode
        self.scale = scale
        self.index = 0
        self.left = [left] if type(left) is str else list(left)
        self.right = [right] if type(right) is str else list(right)
        self.blending = [blending] if type(blending) is float else list(blending)
        self.transforms = [transform] if type(transform) is str else list(transform)
        self.colormaps = [colormap] if type(colormap) is str else list(colormap)
        self.transform_map = {
            'none': functools.partial(self.__no_transform),
            'minmax': functools.partial(self.__minmax_normalization),
            'sum_minmax': functools.partial(self.__sum_minmax_norm)
        }
        self.colorize_map = { "none": lambda x: x.cpu().numpy() }
        self.colorize_map.update(mic.COLORMAPS)
        log.info(f"Exporting blended images @ {self.folder}/[left]_[right]_index.[format].")
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for l, r, b, t, c, f in zip(self.left, self.right, self.blending, 
            self.transforms, self.colormaps, self.formats):
                n = l + "_" + r
                left = tensors[l]
                left = left.cpu().detach().numpy() if left.is_cuda else left.detach().numpy()
                bs = self._save_color(self.folder, 
                    left * b + (1.0 - b) * self.colorize_map[c](
                        self.transform_map[t](tensors[r])
                    ), n, self.index, f, self.scale
                )
        self.index = 0 if self.mode == "overwrite" else self.index + bs

    @staticmethod
    def _save_color(
        path:   str,
        array:  numpy.array,
        key:    str,
        index:  int,
        ext:    str,
        scale:  float,
    ) -> None:
        b, _, __, ___ = array.shape
        for i in range(b):
            img = torch.from_numpy(array[i, ...])
            if scale != 1.0:
               img = torch.nn.functional.interpolate(
                   img.unsqueeze(0), scale_factor=scale, mode='bilinear',
               ).squeeze(0)
            torchvision.utils.save_image(
               img, 
               f'{key}_{index + i}.{ext}'
            )
        return b

    @staticmethod #TODO: refactor these into a common module
    def __no_transform(tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    @staticmethod #TODO: refactor these into a common module
    def __minmax_normalization(tensor: torch.Tensor) -> torch.Tensor:
        b, _, __, ___ = tensor.size()
        min_v = torch.min(tensor.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        max_v = torch.max(tensor.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        return (tensor - min_v) / (max_v - min_v)

    @staticmethod #TODO: refactor these into a common module
    def __sum_minmax_norm(tensor: torch.Tensor) -> torch.Tensor:
        b, _, __, ___ = tensor.size()
        aggregated = torch.sum(tensor, dim=1, keepdim=True)
        min_v = torch.min(aggregated.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        max_v = torch.max(aggregated.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        return (aggregated - min_v) / (max_v - min_v)


import pytorch_lightning
from pytorch_lightning.callbacks import Callback
import pyredner
import typing
import torch

class Redner(Callback): #TODO: need to update all engines to be the same hydra node, this will allow for a single engine config entry for this one
    def __init__(self,
        print_timings:          bool=False,
    ):
        self.timings = print_timings

    def on_fit_start(self, 
        trainer: pytorch_lightning.Trainer, 
        pl_module: pytorch_lightning.LightningModule,
        stage: typing.Optional[str]=None
    ) -> None:
        pyredner.set_print_timing(self.timings)
        pyredner.set_device(pl_module.device)
        pyredner.set_use_gpu(pl_module.device != torch.device('cpu'))

from moai.monads.render.redner.mesh import Silhouette as MeshSilhouette

__all__ = [
    'Redner',
    'MeshSilhouette',
]


from moai.utils.arguments import ensure_choices

import torch
import logging

log = logging.getLogger(__name__)

class L1(torch.nn.Module):
    
    __MODES__ = ['raw', 'ln', 'log']
    __MODES_MAP__ = {
        'raw': lambda e: e,
        'ln': lambda e: torch.log(e + 1.0),
        'log': lambda e: torch.log10(e + 1.0),
    }

    def __init__(self,
        mode:       str='raw', # one of ['raw', 'ln', 'log']
    ):
        super(L1, self).__init__()
        mode = ensure_choices(log, 'mode', mode, L1.__MODES__)
        self.mode = L1.__MODES_MAP__[mode]

    def forward(self,
        pred: torch.Tensor,
        gt: torch.Tensor,
        weights: torch.Tensor=None, # float tensor
        mask: torch.Tensor=None, # byte tensor
    ) -> torch.Tensor:
        l1 = torch.abs(gt - pred) 
        if weights is not None:
            l1 = l1 * weights
        if mask is not None:
            l1 = l1[mask]        
        return self.mode(l1)


from ts.torch_handler.base_handler import BaseHandler
from hydra.experimental import (
    compose,
    initialize,
)
from moai.networks.lightning import Optimizer
from omegaconf.omegaconf import OmegaConf

import hydra.utils as hyu
import toolz
import os
import torch
import zipfile
import yaml
import logging
import typing

log = logging.getLogger(__name__)

__all__ = ['OptimizerServer']

class OptimizerServer(BaseHandler):
    def __init__(self) -> None:
        super().__init__()

    def _extract_files(self):
        with zipfile.ZipFile('conf.zip', 'r') as conf_zip:
            conf_zip.extractall('.')
        with zipfile.ZipFile('src.zip', 'r') as src_zip:
            src_zip.extractall('.')

    def _get_overrides(self):
        if os.path.exists('overrides.yaml'):
            with open('overrides.yaml') as f:
                return yaml.load(f, Loader=yaml.FullLoader)
        else:
            return []

    def initialize(self, context):        
        properties = context.system_properties
        self.map_location = "cuda" if torch.cuda.is_available(
        ) and properties.get("gpu_id") is not None else "cpu"
        self.device = torch.device(
            self.map_location + ":" + str(properties.get("gpu_id"))
            if torch.cuda.is_available() and properties.get("gpu_id") is not None
            else self.map_location
        )
        #NOTE: IMPORTANT!!!! DEBUG WHILE TRAINING ONLY !!!
        self.device = torch.device('cpu')
        log.warning("IMPORTANT: Model explicitly set to CPU mode for debugging purposes!")
        #NOTE: IMPORTANT!!!! DEBUG WHILE TRAINING ONLY !!!
        main_conf = context.manifest['model']['modelName'].replace('_', '/')
        log.info(f"Loading the {main_conf} endpoint.")
        self._extract_files()        
        try:
            with initialize(config_path="conf", job_name=main_conf):
                cfg = compose(
                    config_name=main_conf, 
                    overrides=self._get_overrides()
                )
                self.optimizer = hyu.instantiate(cfg.model)
        except Exception as e:
            log.error(f"An error has occured while loading the model:\n{e}")
        self.optimizer = self.optimizer.to(self.device)
        self.optimizer.eval()
        self.initialized = True
        log.info(f"Model ({type(self.optimizer.model)}) loaded successfully.")
        self.gradient_tolerance = cfg.fitter.gradient_tolerance
        self.relative_tolerance = cfg.fitter.relative_tolerance
        try:
            handler_overrides = None
            if os.path.exists('handler_overrides.yaml'):
                with open('handler_overrides.yaml') as f:
                    handler_overrides = yaml.load(f, Loader=yaml.FullLoader)
                    log.debug(f"Loaded handler overrides:\n{handler_overrides}")
            with initialize(config_path="conf", job_name=f"{main_conf}_preprocess_handlers"):
                cfg = compose(config_name="../pre")
                if handler_overrides is not None and 'preprocess' in handler_overrides.get('handlers', {}):
                    cfg = OmegaConf.merge(cfg, handler_overrides)
                    log.debug(f"Merged handler overrides:\n{cfg}")
                self.preproc = {k: hyu.instantiate(h) for k, h in (toolz.get_in(['handlers', 'preprocess'], cfg) or {}).items()}
            with initialize(config_path="conf", job_name=f"{main_conf}_postprocess_handlers"):
                cfg = compose(config_name="../post")
                if handler_overrides is not None and 'postprocess' in handler_overrides.get('handlers', {}):
                    cfg = OmegaConf.merge(cfg, handler_overrides)
                    log.debug(f"Merged handler overrides:\n{cfg}")
                self.postproc = {k: hyu.instantiate(h) for k, h in (toolz.get_in(['handlers', 'postprocess'], cfg) or {}).items()}
        except Exception as e:
            log.error(f"An error has occured while loading the handlers:\n{e}")

    def preprocess(self, 
        data:   typing.Mapping[str, typing.Any],
    ) -> typing.Dict[str, torch.Tensor]:
        log.debug(f"Preprocessing input:\n{data}")
        tensors = {}
        body = data[0].get('body') or data[0].get('raw')
        for k, p in self.preproc.items():
            tensors = toolz.merge(tensors, p(body, self.device))
        log.debug(f"Tensors: {tensors.keys()}")
        return tensors

    def relative_check(self, 
        prev: torch.Tensor, 
        current: torch.Tensor
    ) -> float:
        relative_change = (prev - current) / max([prev.abs(), current.abs(), 1.0])
        return relative_change <= self.relative_tolerance

    def gradient_check(self, 
        param_groups: typing.Sequence[typing.Dict[str, torch.nn.Parameter]]
    ) -> bool:
        return all(
            p.grad.view(-1).max().abs().item() < self.gradient_tolerance 
            for p in toolz.concat((g['params'] for g in param_groups)) 
            if p.grad is not None
        )

    def is_any_param_nan(self, optimizer: torch.optim.Optimizer) -> bool:
        for pg in optimizer.param_groups:
                for p in pg['params']:
                    if not torch.all(torch.isfinite(p)):
                        return True
        return

    def inference(self, 
        data:       typing.Mapping[str, torch.Tensor],
    ):        
        self.last_loss = None
        self.optimizer.initialize_parameters()                    
        optimizers, schedulers = self.optimizer.configure_optimizers()
        iters = list(toolz.mapcat(lambda o: o.iterations, toolz.unique(optimizers)))
        stages = list(toolz.mapcat(lambda o: o.name, toolz.unique(optimizers)))
        if self.optimizer.mode == 'inference':                
            with torch.no_grad():
                self.optimizer.preprocess(data)
                self.optimizer(data)
                self.optimizer.initialize(data)            
        for i, (optim, iters, stage, sched) in enumerate(zip(
            optimizers, iters, stages, schedulers
        )):
            log.info(f"Optimizing stage: {stage} for {iters} iterations")
            for p in self.optimizer.parameters():
                p.requires_grad_(False)
            for pg in optim.param_groups:
                for p in pg['params']:
                    p.requires_grad_(True)
            def closure():
                self.optimizer.optimizer_zero_grad(
                    epoch=0, batch_idx=0,
                    optimizer=optim, optimizer_idx=i
                )
                self.loss = self.optimizer.training_step(batch=data, 
                    batch_idx=0, optimizer_idx=i
                )['loss']
                self.loss.backward()
                self.optimizer.optimization_step += 1
                data['__moai__']['optimization_step'] = self.optimizer.optimization_step
                return self.loss
            for j in range(iters):
                optim.step(closure=closure)                            
                current_loss = self.loss
                if hasattr(optim, 'assign'):
                    with torch.no_grad():
                        optim.assign(data)
                if (self.last_loss is not None and self.relative_check(
                    self.last_loss, current_loss
                )) or self.gradient_check(optim.param_groups)\
                    or not torch.isfinite(current_loss)\
                    or self.is_any_param_nan(optim):
                        log.warning(f"Optimization stage '{stage}' stopped at iteration {j}/{iters}.")
                        break
            sched.step()
            self.last_loss = current_loss
            self.optimizer.optimization_step = 0
        return data 
       
    def postprocess(self,
        data: typing.Mapping[str, torch.Tensor]
    ) -> typing.Sequence[typing.Any]:
        log.debug(f"Postprocessing outputs:\n{data['__moai__']}")
        outs = []
        for k, p in self.postproc.items():
            outs.append(p(data))
        return outs

    # def handle(self, data, context):
    #     return super(data, context)